#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Core module for Pelix.

Pelix is a Python framework that aims to act as OSGi as much as possible

:author: Thomas Calmant
:copyright: Copyright 2013, isandlaTech
:license: GPLv3
:version: 0.5
:status: Alpha

..

    This file is part of iPOPO.

    iPOPO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    iPOPO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with iPOPO. If not, see <http://www.gnu.org/licenses/>.
"""

# Module version
__version_info__ = (0, 5, 1)
__version__ = ".".join(map(str, __version_info__))

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

# Pelix utility modules
from pelix.utilities import SynchronizedClassMethod, is_string, Deprecated
import pelix.ldapfilter as ldapfilter

# Standard library
import imp
import importlib
import inspect
import logging
import os
import pkgutil
import sys
import threading
import uuid

# ------------------------------------------------------------------------------

ACTIVATOR = "activator"
"""
Name of the module member that will be used as bundle activator.
It must be an object with the following methods:

* start(BundleContext)
* stop(BundleContext)
"""

OBJECTCLASS = "objectClass"
"""
Property containing the list of specifications (strings) provided by a service
"""

SERVICE_ID = "service.id"
"""
Property containing the ID of a service.
This ID is unique in a framework instance.
"""

SERVICE_RANKING = "service.ranking"
"""
Property that indicates the ranking of a service. It is used to sort the
results of methods like get_service_references()
"""

FRAMEWORK_UID = "framework.uid"
"""
Framework instance "unique" identifier. Used in Remote Services to identify
a framework from another.
It can be generated or be forced using the framework initialization properties.
This property is constant during the life of a framework instance.
"""

# Prepare the module logger
_logger = logging.getLogger("pelix.main")

# ------------------------------------------------------------------------------

class BundleException(Exception):
    """
    The base of all framework exceptions
    """
    def __init__(self, content):
        """
        Sets up the exception
        """
        if isinstance(content, Exception):
            Exception.__init__(self, str(content))

        else:
            Exception.__init__(self, content)


class FrameworkException(Exception):
    """
    A framework exception is raised when an error can force the framework to
    stop.
    """
    def __init__(self, message, needs_stop=False):
        """
        Sets up the exception

        :param message: A description of the exception
        :param needs_stop: If True, the framework must be stopped
        """
        Exception.__init__(self, message)
        self.needs_stop = needs_stop

# ------------------------------------------------------------------------------

class Bundle(object):
    """
    Represents a "bundle" in Pelix
    """

    UNINSTALLED = 1
    """ The bundle is uninstalled and may not be used """

    INSTALLED = 2
    """ The bundle is installed but not yet resolved """

    RESOLVED = 4
    """ The bundle is resolved and is able to be started """

    STARTING = 8
    """ The bundle is in the process of starting """

    STOPPING = 16
    """ The bundle is in the process of stopping """

    ACTIVE = 32
    """ The bundle is now running """


    def __init__(self, framework, bundle_id, name, module):
        """
        Sets up the bundle descriptor

        :param framework: The host framework
        :param bundle_id: The ID of the bundle in the host framework
        :param name: The bundle symbolic name
        :param module: The bundle module
        """
        # A reentrant lock for synchronization
        self._lock = threading.RLock()

        # Bundle
        self.__context = BundleContext(framework, self)
        self.__id = bundle_id
        self.__module = module
        self.__name = name

        self.__framework = framework
        self._state = Bundle.RESOLVED

        # Registered services
        self.__registered_services = []
        self.__registration_lock = threading.Lock()


    def __str__(self):
        """
        String representation
        """
        return "Bundle(ID={0}, Name={1})".format(self.__id, self.__name)


    def _fire_bundle_event(self, kind):
        """
        Fires a bundle event of the given kind

        :param kind: Kind of event
        """
        self.__framework._dispatcher.fire_bundle_event(BundleEvent(kind, self))


    def _registered_service(self, registration):
        """
        Bundle is notified by the framework that a service has been registered
        in the name of this bundle.
        
        :param registration: The service registration object
        """
        with self.__registration_lock:
            self.__registered_services.append(registration)


    def _unregistered_service(self, registration):
        """
        Bundle is notified by the framework that a service has been unregistered
        in the name of this bundle.
        
        :param registration: The service registration object
        """
        with self.__registration_lock:
            if registration in self.__registered_services:
                self.__registered_services.remove(registration)


    def get_bundle_context(self):
        """
        Retrieves the bundle context

        :return: The bundle context
        """
        return self.__context


    def get_bundle_id(self):
        """
        Retrieves the bundle ID

        :return: The bundle ID
        """
        return self.__id


    def get_location(self):
        """
        Retrieves the location of this module

        :return: The location of the Pelix module, or an empty string
        """
        return getattr(self.__module, '__file__', "")


    def get_module(self):
        """
        Retrieves the Python module corresponding to the bundle

        :return: The Python module
        """
        return self.__module


    def get_registered_services(self):
        """
        Returns this bundle's ServiceReference list for all services it has
        registered or None if this bundle has no registered services.

        The list is valid at the time of the call to this method, however, as
        the Framework is a very dynamic environment, services can be modified or
        unregistered at any time.
        
        :return: An array of ServiceReference objects or None.
        :raise BundleException: If the bundle has been uninstalled
        """
        if self._state == Bundle.UNINSTALLED:
            raise BundleException("Can't call 'get_registered_services' on an "
                                  "uninstalled bundle")

        return self.__framework._registry.get_bundle_registered_services(self)


    def get_services_in_use(self):
        """
        Returns this bundle's ServiceReference list for all services it is using
        or returns None if this bundle is not using any services.
        A bundle is considered to be using a service if its use count for that
        service is greater than zero.
        
        The list is valid at the time of the call to this method, however, as
        the Framework is a very dynamic environment, services can be modified
        or unregistered at any time.
        
        :return: An array of ServiceReference objects or None.
        :raise BundleException: If the bundle has been uninstalled
        """
        if self._state == Bundle.UNINSTALLED:
            raise BundleException("Can't call 'get_services_in_use' on an "
                                  "uninstalled bundle")

        return self.__framework._registry.get_bundle_imported_services(self)


    def get_state(self):
        """
        Retrieves the bundle state

        :return: The bundle state
        """
        return self._state


    def get_symbolic_name(self):
        """
        Retrieves the bundle symbolic name (its Python module name)

        :return: The bundle symbolic name
        """
        return self.__name


    def get_version(self):
        """
        Retrieves the bundle version, using the ``__version__`` or
        ``__version_info__`` attributes of its module.

        :return: The bundle version, "0.0.0" by default
        """
        # Get the version value
        version = getattr(self.__module, "__version__", None)
        if version:
            return version

        # Convert the __version_info__ entry
        info = getattr(self.__module, "__version_info__", None)
        if info:
            return ".".join(map(str, __version_info__))

        # No version
        return "0.0.0"


    def start(self):
        """
        Starts the bundle. Does nothing if the bundle is already starting or
        active.

        :raise BundleException: The framework is not yet started or the bundle
                                activator failed.
        """
        if self.__framework._state not in (Bundle.STARTING, Bundle.ACTIVE):
            # Framework is not running
            raise BundleException("Framework must be started before its "
                                  "bundles")

        with self._lock:
            if self._state in (Bundle.ACTIVE, Bundle.STARTING):
                # Already started bundle, do nothing
                return

            # Store the bundle current state
            previous_state = self._state

            # Starting...
            self._state = Bundle.STARTING
            self._fire_bundle_event(BundleEvent.STARTING)

            # Call the activator, if any
            activator = getattr(self.__module, ACTIVATOR, None)
            starter = getattr(activator, 'start', None)

            if starter is not None:
                try:
                    # Call the start method
                    starter(self.__context)

                except (FrameworkException, BundleException):
                    # Restore previous state
                    self._state = previous_state

                    # Re-raise directly Pelix exceptions
                    _logger.exception("Pelix error raised by %s while starting",
                                      self.__name)
                    raise

                except Exception as ex:
                    # Restore previous state
                    self._state = previous_state

                    # Raise the error
                    _logger.exception("Error raised by %s while starting",
                                      self.__name)
                    raise BundleException(ex)

            # Bundle is now active
            self._state = Bundle.ACTIVE
            self._fire_bundle_event(BundleEvent.STARTED)


    def stop(self):
        """
        Stops the bundle. Does nothing if the bundle is already stopped.

        :raise BundleException: The bundle activator failed.
        """
        if self._state != Bundle.ACTIVE:
            # Invalid state
            return

        with self._lock:
            # Store the bundle current state
            previous_state = self._state

            # Stopping...
            self._state = Bundle.STOPPING
            self._fire_bundle_event(BundleEvent.STOPPING)

            # Call the activator, if any
            activator = getattr(self.__module, ACTIVATOR, None)
            stopper = getattr(activator, 'stop', None)

            exception = None
            if stopper is not None:
                try:
                    # Call the start method
                    stopper(self.__context)

                except (FrameworkException, BundleException) as ex:
                    # Restore previous state
                    self._state = previous_state

                    # Re-raise directly Pelix exceptions
                    _logger.exception("Pelix error raised by %s while stopping",
                                      self.__name)
                    exception = ex

                except Exception as ex:
                    _logger.exception("Error raised by %s while stopping",
                                      self.__name)
                    # Store the exception (raised after service clean up)
                    exception = BundleException(ex)

            # Intermediate bundle event : activator should have cleaned up
            # everything, but some element could stay (iPOPO components, ...)
            self._fire_bundle_event(BundleEvent.STOPPING_PRECLEAN)

            # Remove remaining services (the hard way)
            self.__unregister_services()

            # Bundle is now stopped and all its services have been unregistered
            self._state = Bundle.RESOLVED
            self._fire_bundle_event(BundleEvent.STOPPED)

        # Raise the exception, if any
        if exception is not None:
            raise exception


    def __unregister_services(self):
        """
        Unregisters all bundle services
        """
        # Copy the services list, as it will be modified during the process
        with self.__registration_lock:
            registered_services = self.__registered_services[:]

        for registration in registered_services:
            try:
                registration.unregister()

            except BundleException:
                # Ignore errors at this level
                pass

        if len(self.__registered_services) != 0:
            _logger.warning("Not all services have been unregistered...")

        with self.__registration_lock:
            # Clear the list, just to be clean
            del self.__registered_services[:]


    def uninstall(self):
        """
        Uninstalls the bundle
        """
        with self._lock:
            if self._state == Bundle.ACTIVE:
                self.stop()

            # Change the bundle state
            self._state = Bundle.UNINSTALLED

            # Call the framework
            self.__framework.uninstall_bundle(self)


    @SynchronizedClassMethod('_lock')
    def update(self):
        """
        Updates the bundle
        """
        # Was it active ?
        restart = self._state == Bundle.ACTIVE

        # Send the update event
        self._fire_bundle_event(BundleEvent.UPDATE_BEGIN)

        try:
            # Stop the bundle
            self.stop()
        except:
            # Something wrong occurred, notify listeners
            self._fire_bundle_event(BundleEvent.UPDATE_FAILED)
            raise

        # Change the source file age
        module_file = getattr(self.__module, "__file__", None)
        can_change = module_file is not None and os.path.isfile(module_file)
        if can_change:
            st = os.stat(module_file)

            # Change modification time to bypass weak time resolution of the
            # underlying file system
            os.utime(module_file, (st.st_atime, st.st_mtime + 1))

        try:
            # Reload the module
            imp.reload(self.__module)

        except SyntaxError as ex:
            # Exception raised in Python 3
            _logger.exception("Error updating %s: %s", self.__name, ex)

        if can_change:
            # Reset times
            os.utime(module_file, (st.st_atime, st.st_mtime))

        if restart:
            try:
                # Re-start the bundle
                self.start()
            except:
                # Something wrong occurred, notify listeners
                self._fire_bundle_event(BundleEvent.UPDATE_FAILED)
                raise

        # Bundle update finished
        self._fire_bundle_event(BundleEvent.UPDATED)

# ------------------------------------------------------------------------------

class Framework(Bundle):
    """
    The Pelix framework (main) class. It must be instantiated using
    FrameworkFactory
    """
    def __init__(self, properties=None):
        """
        Sets up the framework.

        :param properties: The framework properties
        """
        # Framework bundle set up
        Bundle.__init__(self, self, 0, self.get_symbolic_name(),
                        sys.modules[__name__])

        # Framework properties
        if not isinstance(properties, dict):
            self.__properties = {}
        else:
            # Use a copy of the properties, to avoid external changes
            self.__properties = properties.copy()

        # Generate a framework instance UUID, if needed
        framework_uid = self.__properties.get(FRAMEWORK_UID)
        if not framework_uid:
            framework_uid = str(uuid.uuid4())

        # Normalize the UID: it must be a string
        self.__properties[FRAMEWORK_UID] = str(framework_uid)

        # Properties lock
        self.__properties_lock = threading.Lock()

        # Bundles (start at 1, as 0 is reserved for the framework itself)
        self.__next_bundle_id = 1

        # Bundle ID -> Bundle object
        self.__bundles = {}

        # Bundles lock
        self.__bundles_lock = threading.RLock()

        # Event dispatcher
        self._dispatcher = _EventDispatcher()

        # Service registry
        self._registry = _ServiceRegistry(self)
        self.__unregistering_services = {}

        # The wait_for_stop event (initially stopped)
        self._fw_stop_event = threading.Event()
        self._fw_stop_event.set()


    def add_property(self, name, value):
        """
        Adds a property to the framework **if it is not yet set**.

        If the property already exists (same name), then nothing is done.
        Properties can't be updated.

        :param name: The property name
        :param value: The value to set
        :return: True if the property was stored, else False
        """
        with self.__properties_lock:
            if name in self.__properties:
                # Already stored property
                return False

            self.__properties[name] = value
            return True


    def find_service_references(self, clazz=None, ldap_filter=None):
        """
        Finds all services references matching the given filter.

        :param clazz: Class implemented by the service
        :param ldap_filter: Service filter
        :return: A list of found reference, or None
        :raise BundleException: An error occurred looking for service references
        """
        return self._registry.find_service_references(clazz, ldap_filter)


    def get_bundle_by_id(self, bundle_id):
        """
        Retrieves the bundle with the given ID

        :param bundle_id: ID of an installed bundle
        :return: The requested bundle
        :raise BundleException: The ID is invalid
        """
        if bundle_id == 0:
            # "System bundle"
            return self

        with self.__bundles_lock:
            if bundle_id not in self.__bundles:
                raise BundleException("Invalid bundle ID {0}".format(bundle_id))

            return self.__bundles[bundle_id]


    def get_bundle_by_name(self, bundle_name):
        """
        Retrieves the bundle with the given name

        :param bundle_name: Name of the bundle to look for
        :return: The requested bundle, None if not found
        """
        if bundle_name is None:
            # Nothing to do
            return None

        if bundle_name is self.get_symbolic_name():
            # System bundle requested
            return self

        with self.__bundles_lock:
            for bundle in self.__bundles.values():
                if bundle_name == bundle.get_symbolic_name():
                    # Found !
                    return bundle

            # Not found...
            return None


    def get_bundles(self):
        """
        Returns the list of all installed bundles
        
        :return: the list of all installed bundles
        """
        with self.__bundles_lock:
            return list(self.__bundles.values())


    def get_properties(self):
        """
        Retrieves a copy of the stored framework properties.
        """
        with self.__properties_lock:
            return self.__properties.copy()


    def get_property(self, name):
        """
        Retrieves a framework or system property. As framework properties don't
        change while it's running, this method don't need to be protected.

        :param name: The property name
        """
        with self.__properties_lock:
            return self.__properties.get(name, os.getenv(name))


    def get_property_keys(self):
        """
        Returns an array of the keys in the properties of the service

        :return: An array of property keys.
        """
        with self.__properties_lock:
            return tuple(self.__properties.keys())


    def get_service(self, bundle, reference):
        """
        Retrieves the service corresponding to the given reference

        :param bundle: The bundle requiring the service
        :param reference: A service reference
        :return: The requested service
        :raise BundleException: The service could not be found
        :raise TypeError: The argument is not a ServiceReference object
        """
        if not isinstance(bundle, Bundle):
            raise TypeError("First argument must be a Bundle object")

        if not isinstance(reference, ServiceReference):
            raise TypeError("Second argument must be a ServiceReference object")

        if reference in self.__unregistering_services:
            # Unregistering service, just give it
            return self.__unregistering_services[reference]

        return self._registry.get_service(bundle, reference)


    def get_symbolic_name(self):
        """
        Retrieves the framework symbolic name

        :return: Always "pelix.framework"
        """
        return "pelix.framework"


    def install_bundle(self, name, path=None):
        """
        Installs the bundle with the given name
        
        *Note:* Before Pelix 0.5.0, this method returned the ID of the installed
        bundle, instead of the Bundle object. 
        
        **WARNING:** The behavior of the loading process is subject to changes,
        as it does not allow to safely run multiple frameworks in the same
        Python interpreter, as they might share global module values.

        :param name: A bundle name
        :param path: Preferred path to load the module
        :return: The installed Bundle object
        :raise BundleException: Something happened
        """
        with self.__bundles_lock:
            # A bundle can't be installed twice
            for bundle in self.__bundles.values():
                if bundle.get_symbolic_name() == name:
                    _logger.warning('Already installed bundle: %s', name)
                    return bundle

            # Load the module
            try:
                if path:
                    # Use the given path in priority
                    sys.path.insert(0, path)

                if name in sys.modules:
                    # The module has already been loaded
                    module = sys.modules[name]

                else:
                    # Load the module
                    #  __import__(name) -> package level
                    # import_module -> module level
                    module = importlib.import_module(name)

            except ImportError as ex:
                # Error importing the module
                raise BundleException("Error installing bundle {0}: {1}" \
                                      .format(name, ex))

            finally:
                if path:
                    # Clean up the path. The loaded module(s) might
                    # have changed the path content, so do not use an
                    # index
                    sys.path.remove(path)

            # Add the module to sys.modules, just to be sure
            sys.modules[name] = module

            # Compute the bundle ID
            bundle_id = self.__next_bundle_id

            # Prepare the bundle object and its context
            bundle = Bundle(self, bundle_id, name, module)

            # Store the bundle
            self.__bundles[bundle_id] = bundle

            # Update the bundle ID counter
            self.__next_bundle_id += 1

        # Fire the bundle installed event
        event = BundleEvent(BundleEvent.INSTALLED, bundle)
        self._dispatcher.fire_bundle_event(event)

        return bundle


    def install_package(self, path, recursive=False, prefix=None):
        """
        Installs all the modules found in the given package
        
        :param path: Path of the package (folder)
        :param recursive: If True, install the sub-packages too
        :param prefix: (**internal**) Prefix for all found modules
        :return: A 2-tuple, with the list of installed bundles and the list
                 of failed modules names
        :raise ValueError: Invalid path
        """
        if not path:
            raise ValueError("Empty path")

        elif not is_string(path):
            raise ValueError("Path must be a string")

        # Use an absolute path
        path = os.path.abspath(path)
        if not os.path.exists(path):
            raise ValueError("Inexistent path: {0}".format(path))

        # Create a simple visitor
        def visitor(fullname, is_package, module_path):
            # Accept everything in recursive mode, else avoid packages
            return recursive or not is_package

        # Set up the prefix if needed
        if prefix is None:
            prefix = os.path.basename(path)

        bundles = set()
        failed = set()

        with self.__bundles_lock:
            try:
                # Install the package first, resolved from the parent directory
                bundles.add(self.install_bundle(prefix, os.path.dirname(path)))

                # Visit the package
                visited, sub_failed = self.install_visiting(path, visitor,
                                                            prefix)

                # Update the sets
                bundles.update(visited)
                failed.update(sub_failed)

            except BundleException as ex:
                # Error loading the module
                _logger.warning("Error loading package %s: %s", prefix, ex)
                failed.add(prefix)

        return bundles, failed


    def install_visiting(self, path, visitor, prefix=None):
        """
        Installs all the modules found in the given path if they are accepted
        by the visitor.
        
        The visitor must be a callable accepting 3 parameters:
        
           * fullname: The full name of the module
           * is_package: If True, the module is a package
           * module_path: The path to the module file
        
        :param path: Root search path
        :param visitor: The visiting callable
        :param prefix: (**internal**) Prefix for all found modules
        :return: A 2-tuple, with the list of installed bundles and the list
                 of failed modules names
        :raise ValueError: Invalid path or visitor
        """
        # Validate the path
        if not path:
            raise ValueError("Empty path")

        elif not is_string(path):
            raise ValueError("Path must be a string")

        # Validate the visitor
        if visitor is None:
            raise ValueError("No visitor method given")

        # Use an absolute path
        path = os.path.abspath(path)
        if not os.path.exists(path):
            raise ValueError("Inexistent path: {0}".format(path))

        # Set up the prefix if needed
        if prefix is None:
            prefix = os.path.basename(path)

        bundles = set()
        failed = set()

        with self.__bundles_lock:
            # Use an ImpImporter per iteration because, in Python 3,
            # pkgutil.iter_modules() will use a _FileImporter on the second walk
            # in a package which will return nothing
            for name, is_package in pkgutil.ImpImporter(path).iter_modules():
                # Compute the full name of the module
                fullname = '.'.join((prefix, name)) if prefix else name

                try:
                    if visitor(fullname, is_package, path):
                        if is_package:
                            # Install the package
                            bundles.add(self.install_bundle(fullname, path))

                            # Visit the package
                            sub_path = os.path.join(path, name)
                            sub_bundles, sub_failed = self.install_visiting(
                                                                sub_path,
                                                                visitor,
                                                                fullname)
                            bundles.update(sub_bundles)
                            failed.update(sub_failed)

                        else:
                            # Install the bundle
                            bundles.add(self.install_bundle(fullname, path))

                except BundleException as ex:
                    # Error loading the module
                    _logger.warning("Error visiting %s: %s", fullname, ex)

                    # Try the next module
                    failed.add(fullname)
                    continue

        return bundles, failed


    def register_service(self, bundle, clazz, service, properties, send_event):
        """
        Registers a service and calls the listeners

        :param bundle: The bundle registering the service
        :param clazz: Name(s) of the interface(s) implemented by service
        :param properties: Service properties
        :param send_event: If not, doesn't trigger a service registered event
        :return: A ServiceRegistration object
        :raise BundleException: An error occurred while registering the service
        """
        if bundle is None or service is None or not clazz:
            raise BundleException("Invalid registration parameters")

        if not isinstance(properties, dict):
            # Be sure we have a valid dictionary
            properties = {}

        else:
            # Use a copy of the given properties
            properties = properties.copy()

        # Prepare the class specification
        if not isinstance(clazz, list):
            # Make a list from the single class
            clazz = [clazz]

        # Test the list content
        classes = []
        for svc_clazz in clazz:

            if inspect.isclass(svc_clazz):
                # Keep the type name
                svc_clazz = svc_clazz.__name__

            if not svc_clazz or not is_string(svc_clazz):
                # Invalid class name
                raise BundleException("Invalid class name: {0}" \
                                      .format(svc_clazz))

            # Class OK
            classes.append(svc_clazz)

        # Make the service registration
        registration = self._registry.register(bundle, classes, properties,
                                               service)

        # Update the bundle registration information
        bundle._registered_service(registration)

        if send_event:
            # Call the listeners
            event = ServiceEvent(ServiceEvent.REGISTERED,
                                 registration.get_reference())
            self._dispatcher.fire_service_event(event)

        return registration


    @SynchronizedClassMethod('_lock')
    def start(self):
        """
        Starts the framework

        :return: True if the bundle has been started, False if it was already
                 running
        :raise BundleException: A bundle failed to start
        """
        if self._state in (Bundle.STARTING, Bundle.ACTIVE):
            # Already started framework
            return

        # Reset the stop event
        self._fw_stop_event.clear()

        # Starting...
        self._state = Bundle.STARTING
        self._dispatcher.fire_bundle_event(BundleEvent(BundleEvent.STARTING,
                                                       self))

        # Start all registered bundles (use a copy, just in case...)
        for bundle in self.__bundles.copy().values():
            try:
                bundle.start()

            except FrameworkException as ex:
                # Important error
                _logger.exception("Important error starting bundle: %s", bundle)
                if ex.needs_stop:
                    # Stop the framework (has to be in active state)
                    self._state = Bundle.ACTIVE
                    self.stop()
                    return False

            except BundleException:
                # A bundle failed to start : just log
                _logger.exception("Error starting bundle: %s", bundle)

        # Bundle is now active
        self._state = Bundle.ACTIVE
        return True


    @SynchronizedClassMethod('_lock')
    def stop(self, force=False):
        """
        Stops the framework

        :return: True if the framework stopped, False it wasn't running
        """
        if self._state != Bundle.ACTIVE:
            # Invalid state
            return False

        # Stopping...
        self._state = Bundle.STOPPING
        self._dispatcher.fire_bundle_event(BundleEvent(BundleEvent.STOPPING,
                                                       self))

        # Notify listeners that the bundle is stopping
        self._dispatcher.fire_framework_stopping()

        bid = self.__next_bundle_id - 1
        while bid > 0:
            bundle = self.__bundles.get(bid, None)
            bid -= 1

            if bundle is None or bundle.get_state() != Bundle.ACTIVE:
                # Ignore inactive bundle
                continue

            try:
                bundle.stop()

            except Exception as ex:
                # Just log exceptions
                _logger.exception("Error stopping bundle %s: %s",
                                  bundle.get_symbolic_name(), ex)

        # Framework is now stopped
        self._state = Bundle.RESOLVED
        self._dispatcher.fire_bundle_event(BundleEvent(BundleEvent.STOPPED,
                                                       self))

        # All bundles have been stopped, release "wait_for_stop"
        self._fw_stop_event.set()

        # Force the registry clean up
        self._registry.clear()

        return True


    def uninstall(self):
        """
        A framework can't be uninstalled

        :raise BundleException: This method must not be called
        """
        raise BundleException("A framework can't be uninstalled")


    def uninstall_bundle(self, bundle):
        """
        Ends the uninstallation of the given bundle (must be called by Bundle)

        :param bundle: The bundle to uninstall
        :raise BundleException: Invalid bundle
        """
        if bundle is None:
            # Do nothing
            return

        with self.__bundles_lock:
            # Stop the bundle first
            bundle.stop()

            bundle_id = bundle.get_bundle_id()
            if bundle_id not in self.__bundles:
                raise BundleException("Invalid bundle {0}".format(bundle))

            # Notify listeners
            self._dispatcher.fire_bundle_event(BundleEvent(BundleEvent.UNINSTALLED,
                                                           bundle))

            # Remove it from the dictionary
            del self.__bundles[bundle_id]

            # Remove it from the system => avoid unintended behaviors and forces
            # a complete module reload if it is re-installed
            name = bundle.get_symbolic_name()
            if name in sys.modules:
                del sys.modules[name]


    def unregister_service(self, registration):
        """
        Unregisters the given service

        :param registration: A ServiceRegistration to the service to unregister
        :raise BundleException: Invalid reference
        """
        assert isinstance(registration, ServiceRegistration)

        # Get the Service Reference
        reference = registration.get_reference()

        # Remove the service from the registry
        svc_instance = self._registry.unregister(reference)

        # Keep a track of the unregistering reference
        self.__unregistering_services[reference] = svc_instance

        # Call the listeners
        event = ServiceEvent(ServiceEvent.UNREGISTERING, reference)
        self._dispatcher.fire_service_event(event)

        # Update the bundle registration information
        bundle = reference.get_bundle()
        bundle._unregistered_service(registration)

        # Remove the unregistering reference
        del self.__unregistering_services[reference]

        return True


    def update(self):
        """
        Stops and starts the framework, if the framework is active.
        
        :raise BundleException: Something wrong occurred while stopping or 
                                starting the framework.
        """
        with self._lock:
            if self._state == Bundle.ACTIVE:
                self.stop()
                self.start()


    def wait_for_stop(self, timeout=None):
        """
        Waits for the framework to stop. Does nothing if the framework bundle
        is not in ACTIVE state.

        Uses a threading.Condition object

        :param timeout: The maximum time to wait (in seconds)
        :return: True if the framework has stopped, False if the timeout raised
        """
        if self._state != Bundle.ACTIVE:
            # Inactive framework, ignore the call
            return True

        self._fw_stop_event.wait(timeout)

        with self._lock:
            # If the timeout raised, we should be in another state
            return self._state == Bundle.RESOLVED

# ------------------------------------------------------------------------------

class _EventDispatcher(object):
    """
    Simple event dispatcher
    """
    def __init__(self):
        """
        Sets up the dispatcher
        """
        # Bundle listeners
        self.__bnd_listeners = []
        self.__bnd_lock = threading.RLock()

        # Service listeners (listener -> filter)
        self.__svc_listeners = {}
        self.__svc_lock = threading.RLock()

        # Framework stop listeners
        self.__fw_listeners = []
        self.__fw_lock = threading.RLock()


    def clear(self):
        """
        Clears the event dispatcher
        """
        with self.__bnd_lock:
            del self.__bnd_listeners[:]

        with self.__svc_lock:
            self.__svc_listeners.clear()

        with self.__fw_lock:
            del self.__fw_listeners[:]


    def add_bundle_listener(self, listener):
        """
        Adds a bundle listener
        
        :param listener: The bundle listener to register
        :return: True if the listener has been registered, False if it was
                 already known
        :raise BundleException: An invalid listener has been given
        """
        if listener is None or not hasattr(listener, 'bundle_changed'):
            raise BundleException("Invalid bundle listener given")

        with self.__bnd_lock:
            if listener in self.__bnd_listeners:
                _logger.warning("Already known bundle listener '%s'",
                                listener)
                return False

            self.__bnd_listeners.append(listener)
            return True


    def add_framework_listener(self, listener):
        """
        Registers a listener that will be called back right before the framework
        stops.
        
        :param listener: The framework stop listener
        :return: True if the listener has been registered, False if it was
                 already known
        :raise BundleException: An invalid listener has been given
        """
        if listener is None or not hasattr(listener, 'framework_stopping'):
            raise BundleException("Invalid framework listener given")

        with self.__fw_lock:
            if listener in self.__fw_listeners:
                _logger.warning("Already known framework listener '%s'",
                                listener)
                return False

            self.__fw_listeners.append(listener)
            return True


    def add_service_listener(self, listener, ldap_filter=None):
        """
        Registers a service listener

        :param listener: The service listener
        :param ldap_filter: Listener
        :return: True if the listener has been registered, False if it was
                 already known
        :raise BundleException: An invalid listener has been given
        """
        if listener is None or not hasattr(listener, 'service_changed'):
            raise BundleException("Invalid service listener given")

        with self.__svc_lock:
            if listener in self.__svc_listeners:
                _logger.warning("Already known service listener '%s'",
                                listener)
                return False

            try:
                ldap_filter = ldapfilter.get_ldap_filter(ldap_filter)
            except ValueError as ex:
                raise BundleException("Invalid service filter: {0}".format(ex))

            self.__svc_listeners[listener] = ldap_filter
            return True


    def remove_bundle_listener(self, listener):
        """
        Unregisters a bundle listener
        
        :param listener: The bundle listener to unregister
        :return: True if the listener has been unregistered, else False
        """
        with self.__bnd_lock:
            if listener not in self.__bnd_listeners:
                return False

            self.__bnd_listeners.remove(listener)
            return True


    def remove_framework_listener(self, listener):
        """
        Unregisters a framework stop listener
        
        :param listener: The framework listener to unregister
        :return: True if the listener has been unregistered, else False
        """
        with self.__fw_lock:
            if listener not in self.__fw_listeners:
                return False

            self.__fw_listeners.remove(listener)
            return True


    def remove_service_listener(self, listener):
        """
        Unregisters a service listener

        :param listener: The service listener
        :return: True if the listener has been unregistered
        """
        with self.__svc_lock:
            if listener not in self.__svc_listeners:
                return False

            del self.__svc_listeners[listener]
            return True


    def fire_bundle_event(self, event):
        """
        Notifies bundle events listeners of a new event in the calling thread.
        
        :param event: The bundle event
        """
        with self.__bnd_lock:
            # Copy the list of listeners
            listeners = self.__bnd_listeners[:]

        # Call'em all
        for listener in listeners:
            try:
                listener.bundle_changed(event)
            except:
                _logger.exception("Error calling a bundle listener")


    def fire_framework_stopping(self):
        """
        Calls all framework listeners, telling them that the framework is
        stopping
        """
        with self.__fw_lock:
            # Copy the list of listeners
            listeners = self.__fw_listeners[:]

        for listener in listeners:
            try:
                listener.framework_stopping()

            except:
                _logger.exception("An error occurred calling one of the " \
                                  "framework stop listeners")


    def fire_service_event(self, event):
        """
        Notifies service events listeners of a new event in the calling thread.
        
        :param event: The service event
        """
        with self.__svc_lock:
            # Copy the list of listeners
            listeners = self.__svc_listeners.copy()

        # Get the service properties
        properties = event.get_service_reference().get_properties()
        previous = None
        endmatch_event = None
        svc_modified = (event.get_kind() == ServiceEvent.MODIFIED)

        if svc_modified:
            # Modified service event : prepare the end match event
            previous = event.get_previous_properties()
            endmatch_event = ServiceEvent(ServiceEvent.MODIFIED_ENDMATCH,
                                          event.get_service_reference(),
                                          previous)

        # Call'em all
        for listener, ldap_filter in listeners.items():
            # Default event to send : the one we received
            sent_event = event

            # Test if the service properties matches the filter
            if ldap_filter is not None and not ldap_filter.matches(properties):
                # Event doesn't match listener filter...
                if svc_modified and previous is not None \
                and ldap_filter.matches(previous):
                    # ... but previous properties did match
                    sent_event = endmatch_event

                else:
                    # Didn't match before either, ignore it
                    continue

            # Call'em
            try:
                listener.service_changed(sent_event)

            except:
                _logger.exception("Error calling a service listener")

# ------------------------------------------------------------------------------

class _ServiceRegistry(object):
    """
    Service registry for Pelix.
    
    Associates service references to instances and bundles.
    """
    def __init__(self, framework):
        """
        Sets up the registry
        """
        # Associated framework
        self.__framework = framework

        # Next service ID
        self.__next_service_id = 1

        # Service reference -> Service instance
        self.__svc_registry = {}

        # Service reference -> Bundle
        self.__svc_bundle = {}

        # Bundle -> Service references[]
        self.__bundle_svc = {}

        # Bundle -> Service references[]
        self.__bundle_imports = {}

        # Locks
        self.__svc_lock = threading.Lock()


    def clear(self):
        """
        Clears the registry
        """
        with self.__svc_lock:
            self.__svc_registry.clear()
            self.__svc_bundle.clear()
            self.__bundle_svc.clear()
            self.__bundle_imports.clear()


    def register(self, bundle, classes, properties, svc_instance):
        """
        Registers a service.
        
        :param bundle: The bundle that registers the service
        :param classes: The classes implemented by the service
        :param properties: The properties associated to the service
        :param svc_instance: The instance of the service
        :return: The ServiceRegistration object
        """
        with self.__svc_lock:
            # Prepare properties
            service_id = self.__next_service_id
            self.__next_service_id += 1
            properties[OBJECTCLASS] = classes
            properties[SERVICE_ID] = service_id

            # Make the service reference
            svc_ref = ServiceReference(bundle, properties)

            # Make the service registration
            svc_registration = ServiceRegistration(self.__framework, svc_ref,
                                                   properties)

            # Store service information
            self.__svc_registry[svc_ref] = svc_instance
            self.__svc_bundle[svc_ref] = bundle

            # Reverse map, to ease bundle/service association
            self.__bundle_svc.setdefault(bundle, []).append(svc_ref)

            return svc_registration


    def unregister(self, svc_ref):
        """
        Unregisters a service
        
        :param svc_ref: A service reference
        :return: The unregistered service instance
        :raise BundleException: Unknown service reference
        """
        with self.__svc_lock:
            if svc_ref not in self.__svc_registry:
                raise BundleException("Unknown service: {0}".format(svc_ref))

            # Get the owner
            bundle = self.__svc_bundle[svc_ref]

            # Get the service instance
            service = self.__svc_registry[svc_ref]

            # Delete bundle association
            self.__bundle_svc[bundle].remove(svc_ref)
            if len(self.__bundle_svc[bundle]) == 0:
                # Don't keep empty lists
                del self.__bundle_svc[bundle]

            # Delete service information
            del self.__svc_registry[svc_ref]
            del self.__svc_bundle[svc_ref]

            return service


    def find_service_references(self, clazz=None, ldap_filter=None):
        """
        Finds all services references matching the given filter.

        :param clazz: Class implemented by the service
        :param ldap_filter: Service filter
        :return: A list of found reference, or None
        :raise BundleException: An error occurred looking for service references
        """
        with self.__svc_lock:
            if clazz is None and ldap_filter is None:
                # Return a sorted copy of the keys list
                result = list(self.__svc_registry.keys())
                result.sort()
                return result

            if hasattr(clazz, '__name__'):
                # Escape the type name
                clazz = ldapfilter.escape_LDAP(clazz.__name__)

            elif is_string(clazz):
                # Escape the class name
                clazz = ldapfilter.escape_LDAP(clazz)

            if clazz is None:
                # Directly use the given filter
                new_filter = ldap_filter

            elif ldap_filter is None:
                # Make a filter for the object class
                new_filter = "({0}={1})".format(OBJECTCLASS, clazz)

            else:
                # Combine filter with a AND operator
                new_filter = ldapfilter.combine_filters(\
                                        ["({0}={1})".format(OBJECTCLASS, clazz),
                                         ldap_filter])

            # Parse the filter
            try:
                new_filter = ldapfilter.get_ldap_filter(new_filter)

            except ValueError as ex:
                raise BundleException(ex)

            if new_filter is None:
                # Normalized filter is None : return everything
                result = list(self.__svc_registry.keys())

            else:
                # Find a reference that matches
                result = [ref for ref in self.__svc_registry
                          if new_filter.matches(ref.get_properties())]

            if not result:
                # No result found
                return None

            # Sort the results
            result.sort()
            return result


    def get_bundle_imported_services(self, bundle):
        """
        Returns this bundle's ServiceReference list for all services it is using
        or returns None if this bundle is not using any services.
        A bundle is considered to be using a service if its use count for that
        service is greater than zero.
        
        The list is valid at the time of the call to this method, however, as
        the Framework is a very dynamic environment, services can be modified
        or unregistered at any time.
        
        :param bundle: The bundle to look into
        :return: An array of ServiceReference objects or None.
        :raise BundleException: If the bundle has been uninstalled
        """
        with self.__svc_lock:
            return self.__bundle_imports.get(bundle, None)


    def get_bundle_registered_services(self, bundle):
        """
        Retrieves the services registered by the given bundle. Returns None
        if the bundle didn't register any service.
        
        :param bundle: The bundle to look into
        :return: The services registered by the bundle, or None
        """
        with self.__svc_lock:
            return self.__bundle_svc.get(bundle, None)


    def get_service(self, bundle, reference):
        """
        Retrieves the service corresponding to the given reference

        :param bundle: The bundle requiring the service
        :param reference: A service reference
        :return: The requested service
        :raise BundleException: The service could not be found
        """
        with self.__svc_lock:
            # Be sure to have the instance
            if reference in self.__svc_registry:
                service = self.__svc_registry[reference]

                # Indicate the dependency
                imports = self.__bundle_imports.setdefault(bundle, [])
                imports.append(reference)
                reference.used_by(bundle)

                return service

            else:
                # Not found
                raise BundleException("Service not found (reference: {0})" \
                                      .format(reference))


    def unget_service(self, bundle, reference):
        """
        Removes the usage of a service by a bundle

        :param bundle: The bundle that used the service
        :param reference: A service reference
        :return: True if the bundle usage has been removed
        """
        with self.__svc_lock:
            if bundle not in self.__bundle_imports:
                # Unknown bundle
                return False

            imports = self.__bundle_imports[bundle]
            if reference not in imports:
                # Unused reference
                return False

            # Clean up
            imports.remove(reference)
            reference.unused_by(bundle)
            return True


# ------------------------------------------------------------------------------

class BundleContext(object):
    """
    Represents a bundle context
    """
    def __init__(self, framework, bundle):
        """
        Sets up the bundle context

        :param framework: Hosting framework
        :param bundle: The associated bundle
        """
        self.__bundle = bundle
        self.__framework = framework


    def __str__(self):
        """
        String representation
        """
        return "BundleContext({0})".format(self.__bundle)


    def add_bundle_listener(self, listener):
        """
        Registers a bundle listener

        The bundle listener must have a method with the following prototype :

        .. python::

           def bundle_changed(self, bundle_event):
               '''
               :param bundle_event: A BundleEvent object
               '''
               # ...

        :param listener: The bundle listener
        :return: True if the listener has been registered
        """
        return self.__framework._dispatcher.add_bundle_listener(listener)


    def add_framework_stop_listener(self, listener):
        """
        Registers a listener that will be called back right before the framework
        stops

        The framework listener must have a method with the following prototype :

        .. python::

           def framework_stopping(self):
               '''
               No parameter given
               '''
               # ...

        :param listener: The framework stop listener
        :return: True if the listener has been registered
        """
        return self.__framework._dispatcher.add_framework_listener(listener)


    def add_service_listener(self, listener, ldap_filter=None):
        """
        Registers a service listener

        The service listener must have a method with the following prototype :

        .. python::

           def service_changed(self, event):
               '''
               Called by Pelix when some service properties changes

               :param event: A ServiceEvent object
               '''
               # ...

        :param listener: The listener to register
        :param ldap_filter: An LDAP filter on the service properties
        :return: True if the listener has been successfully registered
        """
        return self.__framework._dispatcher.add_service_listener(listener,
                                                                ldap_filter)


    def get_all_service_references(self, clazz, ldap_filter):
        """
        Returns an array of ServiceReference objects.
        The returned array of ServiceReference objects contains services that
        were registered under the specified class and match the specified filter
        expression.
        """
        return self.__framework.find_service_references(clazz, ldap_filter)


    def get_bundle(self, bundle_id=None):
        """
        Retrieves the bundle with the given ID. If no ID is given (None).

        :param bundle_id: A bundle ID
        :return: The requested bundle
        :raise BundleException: The given ID is invalid
        """
        if bundle_id is None:
            # Current bundle
            return self.__bundle

        elif type(bundle_id) is Bundle:
            # Got a bundle (compatibility with older install_bundle())
            bundle_id = bundle_id.get_bundle_id()

        return self.__framework.get_bundle_by_id(bundle_id)


    def get_bundles(self):
        """
        Returns the list of all installed bundles
        
        :return: the list of all installed bundles
        """
        return self.__framework.get_bundles()


    def get_property(self, name):
        """
        Returns the value of a property of the framework, else returns the OS
        environment value.

        :param name: A property name
        """
        return self.__framework.get_property(name)


    def get_service(self, reference):
        """
        Returns the service described with the given reference
        
        :param reference: A ServiceReference object
        :return: The service object itself
        """
        return self.__framework.get_service(self.__bundle, reference)


    def get_service_reference(self, clazz, ldap_filter=None):
        """
        Returns a ServiceReference object for a service that implements and \
        was registered under the specified class

        :param clazz: The class name with which the service was registered.
        :param ldap_filter: A filter on service properties
        :return: A service reference, None if not found
        """
        refs = self.__framework.find_service_references(clazz, ldap_filter)
        if refs is not None and len(refs) > 0:
            return refs[0]

        return None


    def get_service_references(self, clazz, ldap_filter):
        """
        Returns the service references for services that were registered under
        the specified class by this bundle and matching the given filter

        :param ldap_filter: A filter on service properties
        :return: The list of references to the services registered by the
                 calling bundle and matching the filters.
        """
        refs = self.__framework.find_service_references(clazz, ldap_filter)
        for ref in refs:
            if ref.get_bundle() is not self.__bundle:
                refs.remove(ref)

        return refs


    def install_bundle(self, name, path=None):
        """
        Installs the bundle with the given name
        
        *Note:* Before Pelix 0.5.0, this method returned the ID of the installed
        bundle, instead of the Bundle object. 
        
        **WARNING:** The behavior of the loading process is subject to changes,
        as it does not allow to safely run multiple frameworks in the same
        Python interpreter, as they might share global module values.

        :param name: The name of the bundle to install
        :param path: Preferred path to load the module
        :return: The installed Bundle object
        :raise BundleException: Something happened
        """
        return self.__framework.install_bundle(name, path)


    def install_package(self, path, recursive=False):
        """
        Installs all the modules found in the given package
        
        :param path: Path of the package (folder)
        :param recursive: If True, install the sub-packages too
        :return: A 2-tuple, with the list of installed bundles and the list
                 of failed modules names
        :raise ValueError: Invalid path
        """
        return self.__framework.install_package(path, recursive)


    def install_visiting(self, path, visitor):
        """
        Installs all the modules found in the given path if they are accepted
        by the visitor.
        
        The visitor must be a callable accepting 3 parameters:
        
           * fullname: The full name of the module
           * is_package: If True, the module is a package
           * module_path: The path to the module file
        
        :param path: Root search path
        :param visitor: The visiting callable
        :return: A 2-tuple, with the list of installed bundles and the list
                 of failed modules names
        :raise ValueError: Invalid path or visitor
        """
        return self.__framework.install_visiting(path, visitor)


    def register_service(self, clazz, service, properties, send_event=True):
        """
        Registers a service

        :param clazz: Class or Classes (list) implemented by this service
        :param service: The service instance
        :param properties: The services properties (dictionary)
        :param send_event: If not, doesn't trigger a service registered event
        :return: A ServiceRegistration object
        :raise BundleException: An error occurred while registering the service
        """
        return self.__framework.register_service(self.__bundle, clazz,
                                                service, properties, send_event)


    def remove_bundle_listener(self, listener):
        """
        Unregisters a bundle listener

        :param listener: The bundle listener
        :return: True if the listener has been unregistered
        """
        return self.__framework._dispatcher.remove_bundle_listener(listener)


    def remove_framework_stop_listener(self, listener):
        """
        Unregisters a framework stop listener

        :param listener: The framework stop listener
        :return: True if the listener has been unregistered
        """
        return self.__framework._dispatcher.remove_framework_listener(listener)


    def remove_service_listener(self, listener):
        """
        Unregisters a service listener

        :param listener: The service listener
        :return: True if the listener has been unregistered
        """
        return self.__framework._dispatcher.remove_service_listener(listener)


    def unget_service(self, reference):
        """
        Disables a reference to the service
        
        :return: True if the bundle was using this reference, else False
        """
        # Lose the dependency
        return self.__framework._registry.unget_service(self.__bundle,
                                                        reference)


# ------------------------------------------------------------------------------

class ServiceReference(object):
    """
    Represents a reference to a service
    """
    def __init__(self, bundle, properties):
        """
        Sets up the service reference

        :param bundle: The bundle registering the service
        :param properties: The service properties
        :raise BundleException: The properties doesn't contain mandatory entries
        """
        # Check properties
        for mandatory in (SERVICE_ID, OBJECTCLASS):
            if mandatory not in properties:
                raise BundleException(
                            "A Service must at least have a '{0}' entry"\
                            .format(mandatory))

        # Properties lock (used by ServiceRegistration too)
        self._props_lock = threading.RLock()

        # Usage lock
        self.__usage_lock = threading.Lock()

        self.__bundle = bundle
        self.__properties = properties
        self.__using_bundles = []
        self.__service_id = properties[SERVICE_ID]


    def __str__(self):
        """
        String representation
        """
        return "ServiceReference(ID={0}, Bundle={1}, Specs={2})".format(
                                                self.__service_id,
                                                self.__bundle.get_bundle_id(),
                                                self.__properties[OBJECTCLASS])


    def __hash__(self):
        """
        Returns the service hash, i.e. its ID, unique in a framework instance.
        
        :return: The service ID
        """
        return self.__service_id


    def __cmp__(self, other):
        """
        ServiceReference comparison

        See: http://www.osgi.org/javadoc/r4v43/org/osgi/framework/ServiceReference.html#compareTo%28java.lang.Object%29
        """
        if self is other:
            return 0

        if not isinstance(other, ServiceReference):
            # Not comparable => lesser
            return -1

        if self.__service_id == other.__service_id:
            # Same ID, same service
            return 0

        service_rank = int(self.__properties.get(SERVICE_RANKING, 65535))
        other_rank = int(other.__properties.get(SERVICE_RANKING, 65535))

        if service_rank == other_rank:
            # Same rank, ID discriminates (greater ID, lesser reference)
            if self.__service_id > other.__service_id:
                return -1
            else:
                return 1

        elif service_rank < other_rank:
            # Lesser rank value, lesser reference
            return -1

        else:
            return 1


    def __eq__(self, other):
        """
        Equal to other
        """
        if self is other:
            # Same object
            return True

        if not isinstance(other, ServiceReference):
            # Not a service reference
            return False

        return self.__service_id == other.__service_id


    def __ne__(self, other):
        """
        Inequal to other
        """
        return not self.__eq__(other)


    def __ge__(self, other):
        """
        Greater or equal
        """
        return self.__cmp__(other) >= 0


    def __gt__(self, other):
        """
        Greater than other
        """
        return self.__cmp__(other) > 0


    def __le__(self, other):
        """
        Lesser or equal
        """
        return self.__cmp__(other) <= 0


    def __lt__(self, other):
        """
        Lesser than other
        """
        return self.__cmp__(other) < 0


    def get_bundle(self):
        """
        Returns the bundle that registered this service
        
        :return: the bundle that registered this service
        """
        return self.__bundle


    def get_using_bundles(self):
        """
        Returns the list of bundles that use this service
        
        :return: A list of Bundle objects
        """
        return self.__using_bundles


    def get_properties(self):
        """
        Returns a copy of the service properties
        
        :return: A copy of the service properties
        """
        with self._props_lock:
            return self.__properties.copy()


    def get_property(self, name):
        """
        Retrieves the property value for the given name

        :return: The property value, None if not found
        """
        with self._props_lock:
            return self.__properties.get(name, None)


    def get_property_keys(self):
        """
        Returns an array of the keys in the properties of the service

        :return: An array of property keys.
        """
        with self._props_lock:
            return tuple(self.__properties.keys())


    def unused_by(self, bundle):
        """
        Indicates that this reference is not being used anymore by the given
        bundle.
        This method should only be used by the framework.

        :param bundle: A bundle that used this reference
        """
        if bundle is None or bundle is self.__bundle:
            # Ignore
            return

        with self.__usage_lock:
            if bundle in self.__using_bundles:
                self.__using_bundles.remove(bundle)


    def used_by(self, bundle):
        """
        Indicates that this reference is being used by the given bundle.
        This method should only be used by the framework.

        :param bundle: A bundle using this reference
        """
        if bundle is None or bundle is self.__bundle:
            # Ignore
            return

        with self.__usage_lock:
            if bundle not in self.__using_bundles:
                self.__using_bundles.append(bundle)

# ------------------------------------------------------------------------------

class ServiceRegistration(object):
    """
    Represents a service registration object
    """
    def __init__(self, framework, reference, properties):
        """
        Sets up the service registration object

        :param framework: The host framework
        :param reference: A service reference
        :param properties: A reference to the ServiceReference properties
                           dictionary object
        """
        self.__framework = framework
        self.__reference = reference
        self.__properties = properties


    def __str__(self):
        """
        String representation
        """
        return "ServiceRegistration({0})".format(self.__reference)


    def get_reference(self):
        """
        Returns the reference associated to this registration
        
        :return: A ServiceReference object
        """
        return self.__reference


    def set_properties(self, properties):
        """
        Updates the service properties

        :param properties: The new properties
        :raise TypeError: The argument is not a dictionary
        """
        if not isinstance(properties, dict):
            raise TypeError("Waiting for dictionary")

        # Keys that must not be updated
        forbidden_keys = (OBJECTCLASS, SERVICE_ID)

        for forbidden_key in forbidden_keys:
            if forbidden_key in properties:
                del properties[forbidden_key]

        to_delete = []
        for key, value in properties.items():
            if self.__properties.get(key, None) == value:
                # No update
                to_delete.append(key)

        for key in to_delete:
            # Remove unchanged properties
            del properties[key]

        if not properties:
            # Nothing to do
            return

        with self.__reference._props_lock:
            # Update the properties
            previous = self.__properties.copy()
            self.__properties.update(properties)

            # Trigger a new computation in the framework
            event = ServiceEvent(ServiceEvent.MODIFIED, self.__reference,
                                 previous)

            self.__framework._dispatcher.fire_service_event(event)


    def unregister(self):
        """
        Unregisters the service
        """
        self.__framework.unregister_service(self)

# ------------------------------------------------------------------------------

class BundleEvent(object):
    """
    Represents a bundle event
    """

    INSTALLED = 1
    """The bundle has been installed."""

    STARTED = 2
    """The bundle has been started."""

    STARTING = 128
    """The bundle is about to be activated."""

    STOPPED = 4
    """
    The bundle has been stopped. All of its services have been unregistered.
    """

    STOPPING = 256
    """The bundle is about to deactivated."""

    STOPPING_PRECLEAN = 512
    """
    The bundle has been deactivated, but some of its services may still remain.
    """

    UNINSTALLED = 16
    """The bundle has been uninstalled."""

    UPDATED = 8
    """The bundle has been updated. (called after STARTED) """

    UPDATE_BEGIN = 32
    """ The bundle will be updated (called before STOPPING) """

    UPDATE_FAILED = 64
    """ The bundle update has failed. The bundle might be in RESOLVED state """


    def __init__(self, kind, bundle):
        """
        Sets up the event
        """
        self.__kind = kind
        self.__bundle = bundle


    def __str__(self):
        """
        String representation
        """
        return "BundleEvent({0}, {1})".format(self.__kind, self.__bundle)


    def get_bundle(self):
        """
        Retrieves the modified bundle
        """
        return self.__bundle


    def get_kind(self):
        """
        Retrieves the kind of event
        """
        return self.__kind

# ------------------------------------------------------------------------------

class ServiceEvent(object):
    """
    Represents a service event
    """

    REGISTERED = 1
    """ This service has been registered """

    MODIFIED = 2
    """ The properties of a registered service have been modified """

    UNREGISTERING = 4
    """ This service is in the process of being unregistered """

    MODIFIED_ENDMATCH = 8
    """
    The properties of a registered service have been modified and the new
    properties no longer match the listener's filter
    """

    def __init__(self, kind, reference, previous_properties=None):
        """
        Sets up the event

        :param kind: Kind of event
        :param reference: Reference to the modified service
        :param previous_properties: Previous service properties (for MODIFIED
                                    and MODIFIED_ENDMATCH events)
        """
        self.__kind = kind
        self.__reference = reference

        if previous_properties is not None \
        and not isinstance(previous_properties, dict):
            # Accept None or dict() only
            previous_properties = {}

        self.__previous_properties = previous_properties


    def __str__(self):
        """
        String representation
        """
        return "ServiceEvent({0}, {1})".format(self.__kind, self.__reference)


    def get_previous_properties(self):
        """
        Returns the previous values of the service properties, meaningless if
        the the event is not MODIFIED nor MODIFIED_ENDMATCH.
        
        :return: The previous properties of the service
        """
        return self.__previous_properties


    def get_service_reference(self):
        """
        Returns the reference to the service associated to this event
        
        :return: A ServiceReference object
        """
        return self.__reference


    def get_kind(self):
        """
        Returns the kind of service event (see the constants)
        
        :return: the kind of service event
        """
        return self.__kind


    @Deprecated("ServiceEvent: get_type() must be replaced by get_kind()")
    def get_type(self):
        """
        **DEPRECATED:** Use get_kind() instead
        
        Retrieves the kind of service event.
        """
        return self.__kind

# ------------------------------------------------------------------------------

class FrameworkFactory(object):
    """
    A framework factory
    """

    __singleton = None
    """ The framework singleton """

    @classmethod
    def get_framework(cls, properties=None):
        """
        If it doesn't exist yet, creates a framework with the given properties,
        else returns the current framework instance.

        :return: A Pelix instance
        """
        if cls.__singleton is None:
            cls.__singleton = Framework(properties)

        return cls.__singleton


    @classmethod
    def is_framework_running(cls, framework=None):
        """
        Tests if the given framework has been constructed and not deleted.
        If *framework* is None, then the methods returns if at least one
        framework is running.
        
        :param framework: The framework instance to be tested
        :return: True if the framework is running
        """
        if framework is None:
            return cls.__singleton is not None

        else:
            return cls.__singleton == framework


    @classmethod
    def delete_framework(cls, framework):
        """
        Removes the framework singleton

        :return: True on success, else False
        """
        if cls.__singleton is framework:
            # Stop the framework
            try:
                framework.stop()

            except:
                _logger.exception("Error stopping the framework")

            # Uninstall its bundles
            bundles = framework.get_bundles()
            for bundle in bundles:
                try:
                    bundle.uninstall()
                except:
                    _logger.exception("Error uninstalling bundle %s", \
                                      bundle.get_symbolic_name())

            # Clear the event dispatcher
            framework._dispatcher.clear()

            # Clear the singleton
            cls.__singleton = None
            return True

        return False

# ------------------------------------------------------------------------------

def create_framework(bundles, properties=None,
                     auto_start=False, wait_for_stop=False, auto_delete=False):
    """
    Creates a Pelix framework, installs the given bundles and returns its
    instance reference.
    If *auto_start* is True, the framework will be started once all bundles
    will have been installed 
    If *wait_for_stop* is True, the method will return only when the framework
    will have stopped. This requires *auto_start* to be True.
    If *auto_delete* is True, the framework will be deleted once it has stopped,
    and the method will return None.
    This requires *wait_for_stop* and *auto_start* to be True.
    
    :param bundles: Bundles to initially install (shouldn't be empty if
                    *wait_for_stop* is True)
    :param properties: Optional framework properties
    :param auto_start: If True, the framework will be started immediately
    :param wait_for_stop: If True, the method will return only when the
                          framework will have stopped
    :param auto_delete: If True, deletes the framework once it stopped.
    :return: The framework instance
    :raise ValueError: Only one framework can run at a time
    """
    # Test if a framework already exists
    if FrameworkFactory.is_framework_running(None):
        raise ValueError('A framework is already running')

    # Create the framework
    framework = FrameworkFactory.get_framework(properties)

    # Install bundles
    context = framework.get_bundle_context()
    for bundle in bundles:
        context.install_bundle(bundle)

    if auto_start:
        # Automatically start the framework
        framework.start()

        if wait_for_stop:
            # Wait for the framework to stop
            try:
                framework.wait_for_stop(None)

            except KeyboardInterrupt:
                # Stop keyboard interruptions
                if framework.get_state() == Bundle.ACTIVE:
                    framework.stop()

            if auto_delete:
                # Delete the framework
                FrameworkFactory.delete_framework(framework)
                framework = None

    return framework



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Core iPOPO implementation

:author: Thomas Calmant
:copyright: Copyright 2013, isandlaTech
:license: GPLv3
:version: 0.5
:status: Alpha

..

    This file is part of iPOPO.

    iPOPO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    iPOPO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with iPOPO. If not, see <http://www.gnu.org/licenses/>.
"""

# Module version
__version_info__ = (0, 5, 0)
__version__ = ".".join(map(str, __version_info__))

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

from pelix.framework import BundleContext, ServiceEvent, BundleEvent, \
    Bundle, BundleException
from pelix.utilities import add_listener, remove_listener, is_string

import pelix.ipopo.constants as constants
import pelix.framework as pelix
import pelix.ldapfilter as ldapfilter

# ------------------------------------------------------------------------------

import inspect
import logging
import threading

# ------------------------------------------------------------------------------

# Prepare the module logger
_logger = logging.getLogger("ipopo.core")

# ------------------------------------------------------------------------------

class Requirement(object):
    """
    Represents a component requirement
    """
    # The dictionary form fields (filter is a special case)
    __stored_fields__ = ('specifications', 'aggregate', 'optional')

    def __init__(self, specifications, aggregate=False, optional=False,
                 spec_filter=None):
        """
        Sets up the requirement

        :param specifications: The requirement specification (can't be None)
        :param aggregate: If true, this requirement represents a list
        :param optional: If true, this requirement is optional
        :param spec_filter: A filter to select dependencies

        :raise TypeError: A parameter has an invalid type
        :raise ValueError: An error occurred while parsing the filter
        """
        if not isinstance(specifications, (list, tuple)):
            raise TypeError("Specifications must be a list/tuple of strings")

        if not specifications:
            raise ValueError("A specification must be given")

        self.aggregate = aggregate
        self.optional = optional
        self.specifications = specifications

        # Original filter keeper
        self.__original_filter = None

        # Set up the requirement filter (after setting up self.specification)
        self.filter = None
        self.set_filter(spec_filter)


    def __eq__(self, other):
        """
        Equality test
        """
        if other is self:
            # Identity check
            return True

        if not isinstance(other, Requirement):
            # Different types
            return False

        if self.aggregate != other.aggregate or self.optional != other.optional:
            # Different flags
            return False

        if self.filter != other.filter:
            # Different filters (therefore different specifications)
            return False

        return True


    def __ne__(self, other):
        """
        Inequality test
        """
        return not self.__eq__(other)


    def copy(self):
        """
        Returns a copy of this instance

        :return: A copy of this instance
        """
        return Requirement(self.specifications, self.aggregate, self.optional,
                           self.__original_filter)


    @classmethod
    def from_dictionary_form(cls, dictionary):
        """
        Sets up an instance with the given dictionary form

        :param dictionary: The dictionary form
        :return: A configured requirement instance
        :raise ValueError: An attribute is missing in the dictionary form
        :raise TypeError: Invalid form type (only dictionaries are accepted)
        """
        if not isinstance(dictionary, dict):
            raise TypeError("Invalid form type '{0}'".format(
                                                     type(dictionary).__name__))

        if not "specifications" in dictionary:
            raise ValueError("Missing specifications in the dictionary form")

        specs = dictionary["specifications"]
        aggregate = dictionary.get("aggregate", False)
        optional = dictionary.get("optional", False)
        spec_filter = ldapfilter.get_ldap_filter(dictionary.get("filter", None))

        return cls(specs, aggregate, optional, spec_filter)


    def matches(self, properties):
        """
        Tests if the given _StoredInstance matches this requirement

        :param properties: Service properties
        :return: True if the instance matches this requirement
        """
        if properties is None:
            # No properties : invalid service
            return False

        assert isinstance(properties, dict)

        # Properties filter test
        return self.filter.matches(properties)


    def set_filter(self, spec_filter):
        """
        Changes the current filter for the given one

        :param spec_filter: The new requirement filter
        :raise TypeError: Unknown filter type
        """
        if spec_filter is not None and not is_string(spec_filter) \
        and not isinstance(spec_filter,
                           (ldapfilter.LDAPFilter, ldapfilter.LDAPCriteria)):
            # Unknown type
            raise TypeError("Invalid filter type {0}".format(
                                                    type(spec_filter).__name__))

        ldap_criteria = []
        for spec in self.specifications:
            ldap_criteria.append("({0}={1})".format(pelix.OBJECTCLASS,
                                                ldapfilter.escape_LDAP(spec)))

        # Make the filter, escaping the specification name
        ldap_filter = "(|{0})".format("".join(ldap_criteria))

        if spec_filter is not None:
            # Filter given, keep its string form
            self.__original_filter = str(spec_filter)
            ldap_filter = ldapfilter.combine_filters([ldap_filter, spec_filter])

        else:
            # No filter
            self.__original_filter = None

        # Parse the filter
        self.filter = ldapfilter.get_ldap_filter(ldap_filter)


    def to_dictionary_form(self):
        """
        Returns a dictionary form of the current object

        :raise AttributeError: A field to store is missing in the instance
        """
        result = {}
        for field in self.__stored_fields__:
            result[field] = getattr(self, field)

        # Special case: store the original filter
        result['filter'] = self.__original_filter

        return result

# ------------------------------------------------------------------------------

class _RuntimeDependency(object):
    """
    Manages a required dependency field when a component is running
    """
    def __init__(self, stored_instance, field, requirement):
        """
        Sets up the dependency
        
        :param stored_instance: The iPOPO component StoredInstance object
        :param field: The injected field name
        :param requirement: The Requirement describing this dependency
        """
        # The internal state lock
        self._lock = threading.RLock()

        # The iPOPO StoredInstance object
        self._ipopo_instance = stored_instance

        # The bundle context
        self._context = self._ipopo_instance.bundle_context

        # The associated field
        self.field = field

        # The underlying requirement
        self.requirement = requirement

        # Current field value
        self._value = None


    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        
        :return: The removed bindings (list) or None
        """
        self._lock = None
        self._ipopo_instance = None
        self._context = None
        self.requirement = None
        self._value = None
        self.field = None


    def get_bindings(self):
        """
        Retrieves the list of the references to the bound services
        
        :return: A list of ServiceReferences objects
        """
        raise NotImplementedError


    def get_kinds(self):
        """
        Retrieves the kinds of this handler: 'dependency'
        
        :return: the kinds of this handler
        """
        return (constants.HANDLER_DEPENDENCY,)


    def get_value(self):
        """
        Retrieves the value to inject in the component (can be different of
        self._value)
        
        :return: The value to inject
        """
        return self._value


    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        return (self.requirement is not None and self.requirement.optional) \
            or self._value is not None


    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework
        
        :param svc_ref: A service reference
        """
        raise NotImplementedError


    def on_service_departure(self, svc_ref):
        """
        Called when a service has been registered in the framework
        
        :param svc_ref: A service reference
        """
        raise NotImplementedError


    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been registered in the framework
        
        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        """
        raise NotImplementedError


    def service_changed(self, event):
        """
        Called by the framework when a service event occurs
        """
        if self._ipopo_instance is None \
        or not self._ipopo_instance.check_event(event):
            # stop() and clean() may have been called after we have been put
            # inside a listener list copy...
            # or we've been told to ignore this event
            return

        # Call sub-methods
        kind = event.get_kind()
        svc_ref = event.get_service_reference()

        if kind == ServiceEvent.REGISTERED:
            # Service coming
            self.on_service_arrival(svc_ref)

        elif kind in (ServiceEvent.UNREGISTERING,
                      ServiceEvent.MODIFIED_ENDMATCH):
            # Service gone or not matching anymore
            self.on_service_departure(svc_ref)

        elif kind == ServiceEvent.MODIFIED:
            # Modified properties (can be a new injection)
            self.on_service_modify(svc_ref, event.get_previous_properties())


    def start(self):
        """
        Starts the dependency manager
        """
        self._context.add_service_listener(self, self.requirement.filter)


    def stop(self):
        """
        Stops the dependency manager (must be called before clear())
        """
        self._context.remove_service_listener(self)


class _SimpleDependency(_RuntimeDependency):
    """
    Manages a simple dependency field
    """
    def __init__(self, stored_instance, field, requirement):
        """
        Sets up the dependency
        """
        super(_SimpleDependency, self).__init__(stored_instance, field,
                                                requirement)

        # We have only one reference to keep
        self.reference = None


    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        
        :return: The removed bindings (list) or None
        """
        self.reference = None
        super(_SimpleDependency, self).clear()


    def get_bindings(self):
        """
        Retrieves the list of the references to the bound services
        
        :return: A list of ServiceReferences objects
        """
        result = []
        with self._lock:
            if self.reference:
                result.append(self.reference)

        return result


    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework
        
        :param svc_ref: A service reference
        """
        with self._lock:
            if self._value is None:
                # Inject the service
                self.reference = svc_ref
                self._value = self._context.get_service(svc_ref)

                self._ipopo_instance.bind(self, self._value, self.reference)
                return True


    def on_service_departure(self, svc_ref):
        """
        Called when a service has been unregistered from the framework
        
        :param svc_ref: A service reference
        """
        with self._lock:
            if svc_ref is self.reference:
                # Store the current values
                service, reference = self._value, self.reference

                # Clean the instance values
                self._value = None
                self.reference = None

                self._ipopo_instance.unbind(self, service, reference)
                return True


    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been modified in the framework
        
        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        """
        with self._lock:
            if self.reference is None:
                # A previously registered service now matches our filter
                return self.on_service_arrival(svc_ref)

            else:
                # Notify the property modification
                self._ipopo_instance.update(self, self._value, self.reference,
                                            old_properties)


    def stop(self):
        """
        Stops the dependency manager (must be called before clear())
        """
        super(_SimpleDependency, self).stop()
        if self.reference is not None:
            # Use a list
            result = [(self._value, self.reference)]
        else:
            result = None

        return result


    def try_binding(self):
        """
        Searches for the required service if needed
        
        :raise BundleException: Invalid ServiceReference found
        """
        with self._lock:
            if self.reference is not None:
                # Already bound
                return

            # Get all matching services
            ref = self._context.get_service_reference(None,
                                                      self.requirement.filter)
            if ref is not None:
                # Found a service
                self.on_service_arrival(ref)


class _AggregateDependency(_RuntimeDependency):
    """
    Manages an aggregated dependency field
    """
    def __init__(self, stored_instance, field, requirement):
        """
        Sets up the dependency
        """
        super(_AggregateDependency, self).__init__(stored_instance, field,
                                                   requirement)
        # Reference -> Service
        self.services = {}

        # Future injected value
        self._future_value = None


    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        
        :return: The removed bindings (list) or None
        """
        self.services.clear()
        self.services = None

        if self._future_value is not None:
            del self._future_value[:]
            self._future_value = None

        super(_AggregateDependency, self).clear()


    def get_bindings(self):
        """
        Retrieves the list of the references to the bound services
        
        :return: A list of ServiceReferences objects
        """
        with self._lock:
            return list(self.services.keys())


    def get_value(self):
        """
        Retrieves the value to inject in the component (can be different of
        self._value)
        
        :return: The value to inject
        """
        with self._lock:
            # The value field must be a copy of our list
            if self._future_value is not None:
                self._value = self._future_value[:]
            else:
                self._value = None

            return self._value


    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        return (self.requirement is not None and self.requirement.optional) \
            or self._future_value is not None


    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework
        
        :param svc_ref: A service reference
        """
        with self._lock:
            if svc_ref not in self.services:
                # Get the new service
                service = self._context.get_service(svc_ref)

                if self._future_value is None:
                    # First value
                    self._future_value = []

                # Store the information
                self._future_value.append(service)
                self.services[svc_ref] = service

                self._ipopo_instance.bind(self, service, svc_ref)
                return True


    def on_service_departure(self, svc_ref):
        """
        Called when a service has been unregistered from the framework
        
        :param svc_ref: A service reference
        :return: A tuple (service, reference) if the service has been lost,
                 else None
        """
        with self._lock:
            if svc_ref in self.services:
                # Get the service instance
                service = self.services[svc_ref]

                # Clean the instance values
                del self.services[svc_ref]
                self._future_value.remove(service)

                # Nullify the value if needed
                if not self._future_value:
                    self._future_value = None

                self._ipopo_instance.unbind(self, service, svc_ref)
                return True


    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been modified in the framework
        
        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        :return: A tuple (added, (service, reference)) if the dependency has
                 been changed, else None
        """
        with self._lock:
            if svc_ref not in self.services:
                # A previously registered service now matches our filter
                return self.on_service_arrival(svc_ref)

            else:
                # Notify the property modification
                self._ipopo_instance.update(self, self.services[svc_ref],
                                            svc_ref, old_properties)


    def stop(self):
        """
        Stops the dependency manager (must be called before clear())
        """
        super(_AggregateDependency, self).stop()

        if self.services:
            results = [(service, reference)
                       for reference, service in self.services.items()]

        else:
            results = None

        return results


    def try_binding(self):
        """
        Searches for the required service if needed
        
        :raise BundleException: Invalid ServiceReference found
        """
        with self._lock:
            # Get all matching services
            refs = self._context.get_all_service_references(None,
                                                        self.requirement.filter)
            if not refs:
                # No match found
                return

            # Filter found references
            refs = [reference for reference in refs
                    if reference not in self.services]
            if not refs:
                # No new match found
                return

            results = []
            try:
                # Bind all new reference
                for reference in refs:
                    added = self.on_service_arrival(reference)
                    if added:
                        results.append(reference)

            except BundleException as ex:
                _logger.debug("Error binding multiple references: %s", ex)

                # Undo what has just been done, ignoring errors
                for reference in results:
                    try:
                        self.on_service_departure(reference)

                    except BundleException as ex2:
                        _logger.debug("Error cleaning up: %s", ex2)

                del results[:]
                raise

# ------------------------------------------------------------------------------

class _ServiceRegistrationHandler(object):
    """
    Handles the registration of a service provided by a component
    """
    def __init__(self, specifications, controller_name, ipopo_instance):
        """
        Sets up the handler
        
        :param specifications: The service specifications
        :param controller_name: Name of the associated service controller
                                (can be None)
        :param ipopo_instance: The iPOPO component StoredInstance
        """
        self.specifications = specifications
        self._ipopo_instance = ipopo_instance

        self.__controller = controller_name
        # Controller is "on" by default
        self.__controller_on = True
        self.__validated = False

        # The ServiceRegistration and ServiceReference objects
        self._registration = None
        self._svc_reference = None


    def check_event(self, svc_event):
        """
        Tests if the given service event corresponds to the registered service
        
        :param svc_event: A service event
        :return: True if the given event references the provided service 
        """
        return self._svc_reference is not svc_event.get_service_reference()


    def get_kinds(self):
        """
        Retrieves the kinds of this handler: 'service_provider'
        
        :return: the kinds of this handler
        """
        return (constants.HANDLER_SERVICE_PROVIDER,)


    def get_service_reference(self):
        """
        Retrieves the reference of the provided service
        
        :return: A ServiceReference object
        """
        return self._svc_reference


    def on_controller_change(self, name, value):
        """
        Called by the instance manager when a controller value has been modified
        
        :param name: The name of the controller
        :param value: The new value of the controller
        """
        if self.__controller != name:
            # Nothing to do
            return

        # Update the controller value
        self.__controller_on = value
        if value:
            # Controller switched to "ON"
            self._register_service()

        else:
            # Controller switched to "OFF"
            self._unregister_service()


    def on_property_change(self, name, old_value, new_value):
        """
        Called by the instance manager when a component property is modified

        :param name: The changed property name
        :param old_value: The previous property value
        :param new_value: The new property value
        """
        if self._registration is not None:
            # use the registration to trigger the service event
            self._registration.set_properties({name: new_value})


    def post_validate(self):
        """
        Called by the instance manager once the component has been validated
        """
        # Update the validation flag
        self.__validated = True
        self._register_service()


    def pre_invalidate(self):
        """
        Called by the instance manager before the component is invalidated
        """
        # Update the validation flag
        self.__validated = False

        # Force service unregistration
        self._unregister_service()


    def _register_service(self):
        """
        Registers the provided service, if possible
        """
        if self._registration is None and self.specifications \
        and self.__validated and self.__controller_on:
            # Use a copy of component properties
            properties = self._ipopo_instance.context.properties.copy()
            bundle_context = self._ipopo_instance.bundle_context

            # Register the service
            self._registration = bundle_context.register_service(
                                            self.specifications,
                                            self._ipopo_instance.instance,
                                            properties)
            self._svc_reference = self._registration.get_reference()


    def _unregister_service(self):
        """
        Unregisters the provided service, if needed
        """
        if self._registration is not None:
            # Ignore error
            try:
                self._registration.unregister()

            except BundleException as ex:
                # Only log the error at this level
                _logger.error("Error unregistering a component service: %s", ex)

            self._registration = None
            self._svc_reference = None

# ------------------------------------------------------------------------------

class FactoryContext(object):
    """
    Represents the data stored in a component factory (class)
    """

    __basic_fields = ('callbacks', 'field_callbacks', 'name', 'properties',
                      'properties_fields', 'provides')

    def __init__(self):
        """
        Sets up the factory context
        """
        # Factory bundle context
        self.bundle_context = None

        # Callbacks : Kind -> callback method
        self.callbacks = {}

        # Field callbacks: Field -> {Kind -> Callback}
        self.field_callbacks = {}

        # The factory name
        self.name = None

        # Properties : Name -> Value
        self.properties = {}

        # Properties fields : Field name -> Property name
        self.properties_fields = {}

        # Provided specifications:
        # Array of tuples (specifications(arrays of strings), controller name)
        self.provides = []

        # Requirements : Field name -> Requirement object
        self.requirements = {}


    def __eq__(self, other):
        """
        Equality test
        """
        if other is self:
            # Identity
            return True

        if not isinstance(other, FactoryContext):
            # Different types
            return False

        # Do not compare the bundle context, as it must not be stored
        for field in ('name', 'callbacks', 'field_callbacks', 'properties',
                      'properties_fields', 'requirements'):
            # Comparable fields
            if getattr(self, field, None) != getattr(other, field, None):
                return False

        # Treat the list differently
        if len(self.provides) != len(other.provides):
            return False

        for provided in self.provides:
            if provided not in other.provides:
                # Missing a provided service
                return False

        return True

    def __ne__(self, other):
        """
        Inequality test
        """
        return not self.__eq__(other)


    def copy(self):
        """
        Returns a copy of the current FactoryContext instance
        """
        context = FactoryContext()

        direct = ("bundle_context", "name")
        copied = ("callbacks", "field_callbacks", "properties",
                  "properties_fields", "requirements")
        lists = ("provides",)

        # Direct copy of primitive values
        for entry in direct:
            setattr(context, entry, getattr(self, entry))

        # Copy "complex" values
        for entry in copied:
            value = getattr(self, entry)
            if value is not None:
                value = value.copy()

            setattr(context, entry, value)

        # Copy lists
        for entry in lists:
            value = getattr(self, entry)
            if value is not None:
                value = value[:]

            setattr(context, entry, value)

        return context


    @classmethod
    def from_dictionary_form(cls, dictionary):
        """
        Sets up this instance with the given dictionary form

        :param dictionary: The dictionary form
        :raise ValueError: An attribute is missing in the dictionary form
        :raise TypeError: Invalid form type (only dictionaries are accepted)
        """
        if not isinstance(dictionary, dict):
            raise TypeError("Invalid form type '{0}'".format(
                                                    type(dictionary).__name__))

        # Prepare the instance, initializing it
        instance = cls()

        # Basic fields
        for field in cls.__basic_fields:
            if field not in dictionary:
                raise ValueError("Incomplete dictionary form: missing {0}" \
                                 .format(field))

            setattr(instance, field, dictionary[field])

        # Requirements field
        if 'requirements' not in dictionary:
            raise ValueError("Incomplete dictionary form: missing " \
                             "'requirements'")

        requirements = dictionary['requirements']
        if not isinstance(requirements, dict):
            raise TypeError("Only dictionaries are handled for 'requirements'")

        for field, requirement_dict in requirements.items():
            instance.requirements[field] = Requirement.from_dictionary_form(\
                                                            requirement_dict)

        return instance


    def set_bundle_context(self, bundle_context):
        """
        Sets up the bundle context associated to this factory context

        :param bundle_context: The factory bundle context
        """
        if self.bundle_context is None:
            assert isinstance(bundle_context, BundleContext)
            self.bundle_context = bundle_context


    def to_dictionary_form(self):
        """
        Returns a dictionary form of the current object

        :raise AttributeError: A field to store in missing in the instance
        """
        result = {}

        # Fields with standard Python types (no conversion needed)
        for entry in self.__basic_fields:
            result[entry] = getattr(self, entry)

        # Requirements field
        requirements = {}
        for field, requirement in self.requirements.items():
            requirements[field] = requirement.to_dictionary_form()

        result['requirements'] = requirements
        return result


class ComponentContext(object):
    """
    Represents the data stored in a component instance
    """

    # Try to reduce memory footprint (stored __instances)
    __slots__ = ('factory_context', 'name', 'properties', 'requirements')

    def __init__(self, factory_context, name, properties):
        """
        Sets up the context

        :param factory_context: The parent factory context
        :param properties: The component properties
        """
        assert isinstance(factory_context, FactoryContext)
        assert isinstance(properties, dict)

        self.factory_context = factory_context
        self.name = name

        # Force the instance name property
        properties[constants.IPOPO_INSTANCE_NAME] = name

        self.properties = factory_context.properties.copy()
        self.properties.update(properties)

        requires_filters = self.properties.get(\
                                        constants.IPOPO_REQUIRES_FILTERS, None)

        if not requires_filters or not isinstance(requires_filters, dict):
            # No explicit filter configured
            self.requirements = factory_context.requirements

        else:
            # We need to change a part of the requirements
            self.requirements = {}
            for field, requirement in factory_context.requirements.items():

                if field not in requires_filters:
                    # No information for this one, keep the factory requirement
                    self.requirements[field] = requirement

                else:
                    try:
                        # Use a copy of the requirement
                        requirement_copy = requirement.copy()
                        requirement_copy.set_filter(requires_filters[field])

                        self.requirements[field] = requirement_copy

                    except (TypeError, ValueError):
                        # Invalid filter, use the factory requirement
                        self.requirements[field] = requirement


    def get_bundle_context(self):
        """
        Retrieves the bundle context

        :return: The component bundle context
        """
        return self.factory_context.bundle_context


    def get_callback(self, event):
        """
        Retrieves the registered method for the given event. Returns None if not
        found

        :param event: A component life cycle event
        :return: The callback associated to the given event
        """
        return self.factory_context.callbacks.get(event, None)


    def get_field_callback(self, field, event):
        """
        Retrieves the registered method for the given event. Returns None if not
        found

        :param field: Name of the dependency field
        :param event: A component life cycle event
        :return: The callback associated to the given event
        """
        return self.factory_context.field_callbacks.get(field,
                                                        {}).get(event, None)


    def get_factory_name(self):
        """
        Retrieves the component factory name

        :return: The component factory name
        """
        return self.factory_context.name


    def get_provides(self):
        """
        Retrieves the services provided by this component.
        Returns an array containing arrays of specifications.

        :return: An array of tuples (specifications, controller)
        """
        return self.factory_context.provides



class _StoredInstance(object):
    """
    Represents a component instance
    """
    # Try to reduce memory footprint (stored instances)
    __slots__ = ('bundle_context', 'context', 'factory_name', 'instance',
                 'name', 'state', '_controllers_state', '_handlers',
                 '_ipopo_service', '_lock')

    INVALID = 0
    """ This component has been invalidated """

    VALID = 1
    """ This component has been validated """

    KILLED = 2
    """ This component has been killed """

    VALIDATING = 3
    """ This component is currently validating """

    def __init__(self, ipopo_service, context, instance):
        """
        Sets up the instance object

        :param ipopo_service: The iPOPO service that instantiated this component
        :param context: The component context
        :param instance: The component instance
        """
        assert isinstance(context, ComponentContext)

        # The lock
        self._lock = threading.RLock()

        # The iPOPO service
        self._ipopo_service = ipopo_service

        # Component context
        self.context = context

        # The instance name
        self.name = self.context.name

        # Factory name
        self.factory_name = self.context.get_factory_name()

        # Component instance
        self.instance = instance

        # Set the instance state
        self.state = _StoredInstance.INVALID

        # Store the bundle context
        self.bundle_context = self.context.get_bundle_context()

        # The controllers state dictionary
        self._controllers_state = {}

        # Handlers...
        self._handlers = {}

        # The provided services handlers
        for specs, controller in self.context.get_provides():
            handler = _ServiceRegistrationHandler(specs, controller, self)

            for kind in handler.get_kinds():
                self._handlers.setdefault(kind, []).append(handler)

        # The runtime dependency handlers
        for field, requirement in context.requirements.items():
            if requirement.aggregate:
                handler = _AggregateDependency(self, field, requirement)
            else:
                handler = _SimpleDependency(self, field, requirement)

            for kind in handler.get_kinds():
                self._handlers.setdefault(kind, []).append(handler)


    def __repr__(self):
        """
        String representation
        """
        return self.__str__()


    def __str__(self):
        """
        String representation
        """
        return "StoredInstance(Name={0}, State={1})".format(self.name,
                                                            self.state)


    def check_event(self, event):
        """
        Tests if the given service event must be handled or ignored, based
        on the state of the iPOPO service and on the content of the event.
        
        :param event: A service event
        :return: True if the event can be handled, False if it must be ignored
        """
        with self._lock:
            if self.state == _StoredInstance.KILLED:
                # This call may have been blocked by the internal state lock,
                # ignore it
                return False

            return self.__safe_handlers_callback('check_event', event)


    def bind(self, dependency, svc, svc_ref):
        """
        Called by a dependency manager to inject a new service and update the
        component life cycle.
        """
        with self._lock:
            self.__set_binding(dependency, svc, svc_ref)
            self.check_lifecycle()


    def update(self, dependency, svc, svc_ref, old_properties):
        """
        Called by a dependency manager when the properties of an injected
        dependency have been updated.
        """
        with self._lock:
            self.__update_binding(dependency, svc, svc_ref, old_properties)
            self.check_lifecycle()


    def unbind(self, dependency, svc, svc_ref):
        """
        Called by a dependency manager to remove an injected service and to
        update the component life cycle.
        """
        with self._lock:
            # Invalidate first (if needed)
            self.check_lifecycle()

            # Call unbind() and remove the injection
            self.__unset_binding(dependency, svc, svc_ref)

            # Try a new configuration
            if self.update_bindings():
                self.check_lifecycle()


    def get_controller_state(self, name):
        """
        Retrieves the state of the controller with the given name
        
        :param name: The name of the controller
        :return: The value of the controller
        :raise KeyError: No value associated to this controller
        """
        return self._controllers_state[name]


    def set_controller_state(self, name, value):
        """
        Sets the state of the controller with the given name
        
        :param name: The name of the controller
        :param value: The new value of the controller
        """
        with self._lock:
            self._controllers_state[name] = value
            self.__safe_handlers_callback('on_controller_change', name, value)


    def update_property(self, name, old_value, new_value):
        """
        Handles a property changed event

        :param name: The changed property name
        :param old_value: The previous property value
        :param new_value: The new property value
        """
        with self._lock:
            self.__safe_handlers_callback('on_property_change', name, old_value,
                                          new_value)


    def get_handlers(self, kind=None):
        """
        Retrieves the handlers of the given kind. If kind is None, all handlers
        are returned.
        
        :param kind: The kind of the handlers to return
        :return: A list of handlers, or an empty list
        """
        with self._lock:
            if kind is not None:
                return self._handlers.get(kind, [])

            # Prepare the list of handlers to call
            result = []
            for handlers_list in self._handlers.values():
                for handler in handlers_list:
                    if handler not in result:
                        result.append(handler)

            return result


    def check_lifecycle(self):
        """
        Tests if the state of the component must be updated, based on its own
        state and on the state of its dependencies
        """
        with self._lock:
            # Validation flags
            was_valid = (self.state == _StoredInstance.VALID)
            can_validate = self.state not in (_StoredInstance.VALIDATING,
                                              _StoredInstance.VALID)

            # Test the validity of all handlers
            handlers_valid = self.__safe_handlers_callback('is_valid',
                                                           break_on_false=True)

            # A dependency is missing
            if was_valid and not handlers_valid:
                self.invalidate(True)

            # We're all good
            elif can_validate and handlers_valid \
            and self._ipopo_service.running:
                self.validate(True)


    def update_bindings(self):
        """
        Updates the bindings of the given component

        :return: True if the component can be validated
        """
        with self._lock:
            all_valid = True
            for handler in self._handlers.get(constants.HANDLER_DEPENDENCY, []):
                # Try to bind
                self.__safe_handler_callback(handler, 'try_binding')

                # Update the validity flag
                all_valid &= self.__safe_handler_callback(handler, 'is_valid',
                                                          only_boolean=True,
                                                          none_as_true=True)

            return all_valid


    def start(self):
        """
        Starts the handlers
        """
        with self._lock:
            self.__safe_handlers_callback('start')


    def invalidate(self, callback=True):
        """
        Applies the component invalidation.

        :param callback: If True, call back the component before the
                         invalidation
        """
        with self._lock:
            if self.state != _StoredInstance.VALID:
                # Instance is not running...
                return

            # Change the state
            self.state = _StoredInstance.INVALID

            # Call the handlers
            self.__safe_handlers_callback('pre_invalidate')

            # Call the component
            if callback:
                self.__safe_callback(constants.IPOPO_CALLBACK_INVALIDATE,
                                   self.bundle_context)

                # Trigger an "Invalidated" event
                self._ipopo_service._fire_ipopo_event(
                                              constants.IPopoEvent.INVALIDATED,
                                              self.factory_name, self.name)

            # Call the handlers
            self.__safe_handlers_callback('post_invalidate')


    def kill(self):
        """
        This instance is killed : invalidate it if needed, clean up all members

        When this method is called, this _StoredInstance object must have
        been removed from the registry
        """
        with self._lock:
            # Already dead...
            if self.state == _StoredInstance.KILLED:
                return

            # Unregister from service events
            self.bundle_context.remove_service_listener(self)

            try:
                self.invalidate(True)

            except:
                _logger.exception("%s: Error invalidating the instance",
                                  self.name)

            # Now that we are nearly clean, be sure we were in a good registry
            # state
            assert not self._ipopo_service.is_registered_instance(self.name)

            # Stop all handlers (can tell to unset a binding)
            for handler in self.get_handlers():
                results = self.__safe_handler_callback(handler, 'stop')
                if results:
                    try:
                        for binding in results:
                            self.__unset_binding(handler, binding[0],
                                                 binding[1])

                    except Exception as ex:
                        _logger.exception("Error stopping handler '%s': %s",
                                          handler, ex)

            # Call the handlers
            self.__safe_handlers_callback('clear')

            # Change the state
            self.state = _StoredInstance.KILLED

            # Trigger the event
            self._ipopo_service._fire_ipopo_event(constants.IPopoEvent.KILLED,
                                                  self.factory_name, self.name)

            # Clean up members
            self._handlers.clear()
            self._handlers = None
            self.context = None
            self.instance = None
            self._ipopo_service = None


    def validate(self, safe_callback=True):
        """
        Ends the component validation, registering services

        :param safe_callback: If True, calls the component validation callback
        :raise RuntimeError: You try to awake a dead component
        """
        with self._lock:
            if self.state in (_StoredInstance.VALID,
                              _StoredInstance.VALIDATING):
                # No work to do (yet)
                return

            if self.state == _StoredInstance.KILLED:
                raise RuntimeError("{0}: Zombies !".format(self.name))

            # Call the handlers
            self.__safe_handlers_callback('pre_validate')

            if safe_callback:
                # Safe call back needed and not yet passed
                self.state = _StoredInstance.VALIDATING
                if not self.__safe_callback(constants.IPOPO_CALLBACK_VALIDATE,
                                            self.bundle_context):
                    # Stop there if the callback failed
                    self.invalidate(True)
                    return

            # All good
            self.state = _StoredInstance.VALID

            # Call the handlers
            self.__safe_handlers_callback('post_validate')

            # We may have caused a framework error, so check if iPOPO is active
            if self._ipopo_service is not None:
                # Trigger the iPOPO event (after the service _registration)
                self._ipopo_service._fire_ipopo_event(
                                              constants.IPopoEvent.VALIDATED,
                                              self.factory_name, self.name)


    def __callback(self, event, *args, **kwargs):
        """
        Calls the registered method in the component for the given event

        :param event: An event (IPOPO_CALLBACK_VALIDATE, ...)
        :return: The callback result, or None
        :raise Exception: Something went wrong
        """
        with self._lock:
            comp_callback = self.context.get_callback(event)
            if not comp_callback:
                # No registered callback
                return True

            # Call it
            result = comp_callback(self.instance, *args, **kwargs)
            if result is None:
                # Special case, if the call back returns nothing
                return True

            return result


    def __field_callback(self, field, event, *args, **kwargs):
        """
        Calls the registered method in the component for the given field event

        :param field: A field name
        :param event: An event (IPOPO_CALLBACK_VALIDATE, ...)
        :return: The callback result, or None
        :raise Exception: Something went wrong
        """
        with self._lock:
            comp_callback = self.context.get_field_callback(field, event)
            if not comp_callback:
                # No registered callback
                return True

            # Call it
            result = comp_callback(self.instance, field, *args, **kwargs)
            if result is None:
                # Special case, if the call back returns nothing
                return True

            return result


    def __safe_callback(self, event, *args, **kwargs):
        """
        Calls the registered method in the component for the given event,
        ignoring raised exceptions

        :param event: An event (IPOPO_CALLBACK_VALIDATE, ...)
        :return: The callback result, or None
        """
        with self._lock:
            if self.state == _StoredInstance.KILLED:
                # Invalid state
                return None

            try:
                return self.__callback(event, *args, **kwargs)

            except pelix.FrameworkException as ex:
                # Important error
                _logger.exception("Critical error calling back %s: %s",
                                  self.name, ex)

                # Kill the component
                self._ipopo_service.kill(self.name)

                if ex.needs_stop:
                    # Framework must be stopped...
                    _logger.error("%s said that the Framework must be stopped.",
                                  self.name)
                    self.bundle_context.get_bundle(0).stop()
                return False

            except:
                _logger.exception("Component '{0}' : error calling callback " \
                                  "method for event {1}".format(self.name,
                                                                event))
                return False


    def __safe_field_callback(self, field, event, *args, **kwargs):
        """
        Calls the registered method in the component for the given event,
        ignoring raised exceptions

        :param field: Name of the modified field
        :param event: A field event (IPOPO_CALLBACK_BIND_FIELD, ...)
        :return: The callback result, or None
        """
        with self._lock:
            if self.state == _StoredInstance.KILLED:
                # Invalid state
                return None

            try:
                return self.__field_callback(field, event, *args, **kwargs)

            except pelix.FrameworkException as ex:
                # Important error
                _logger.exception("Critical error calling back %s: %s",
                                  self.name, ex)

                # Kill the component
                self._ipopo_service.kill(self.name)

                if ex.needs_stop:
                    # Framework must be stopped...
                    _logger.error("%s said that the Framework must be stopped.",
                                  self.name)
                    self.bundle_context.get_bundle(0).stop()
                return False

            except:
                _logger.exception("Component '{0}' : error calling callback " \
                                  "method for event {1}".format(self.name,
                                                                event))
                return False


    def __safe_handler_callback(self, handler, method_name, *args, **kwargs):
        """
        Calls the given method with the given arguments in the given handler.
        Logs exceptions, but doesn't propagate them.
        
        Special arguments can be given in kwargs:
        
        * 'none_as_true': If set to True and the method returned None or doesn't
                          exist, the result is considered as True.
                          If set to False, None result is kept as is.
                          Default is False.
        * 'only_boolean': If True, the result can only be True or False, else
                          the result is the value returned by the method.
                          Default is False.
        
        :param handler: The handler to call
        :param method_name: The name of the method to call
        :param args: List of arguments for the method to call
        :param kwargs: Dictionary of arguments for the method to call and to
                       control the call
        :return: The method result, or None on error
        """
        if handler is None or method_name is None:
            return None

        # Behavior flags
        only_boolean = kwargs.pop('only_boolean', False)
        none_as_true = kwargs.pop('none_as_true', False)

        # Get the method for each handler
        result = None
        method = getattr(handler, method_name, None)
        if method is not None:
            try:
                # Call it
                result = method(*args, **kwargs)

            except Exception as ex:
                # Log errors
                _logger.exception("Error calling handler '%s': %s", handler, ex)

        if result is None and none_as_true:
            # Consider None (nothing returned) as True
            result = True

        if only_boolean:
            # Convert to a boolean result
            result = only_boolean and result

        return result


    def __safe_handlers_callback(self, method_name, *args, **kwargs):
        """
        Calls the given method with the given arguments in all handlers.
        Logs exceptions, but doesn't propagate them.
        Methods called in handlers must return None, True or False.
        
        Special parameters can be given in kwargs:
        
        * 'exception_as_error': if it is set to True and an exception is raised
          by a handler, then this method will return False. By default, this
          flag is set to False and exceptions are ignored.
        * 'break_on_false': if it set to True, the loop calling the handler
          will stop after an handler returned False. By default, this flag
          is set to False, and all handlers are called.
        
        :param method_name: Name of the method to call
        :param args: List of arguments for the method to call
        :param kwargs: Dictionary of arguments for the method to call and the
                       behavior of the call
        :return: True if all handlers returned True (or None), else False
        """
        with self._lock:
            if self.state == _StoredInstance.KILLED:
                # Nothing to do
                return False

            # Behavior flags
            exception_as_error = kwargs.pop('exception_as_error', False)
            break_on_false = kwargs.pop('break_on_false', False)

            result = True
            for handler in self.get_handlers():
                # Get the method for each handler
                method = getattr(handler, method_name, None)
                if method is not None:
                    try:
                        # Call it
                        res = method(*args, **kwargs)
                        if res is not None and not res:
                            # Ignore 'None' results
                            result = False

                    except Exception as ex:
                        # Log errors
                        _logger.exception("Error calling handler '%s': %s",
                                          handler, ex)

                        # We can consider exceptions as errors or ignore them
                        result &= not exception_as_error

                    if not handler and break_on_false:
                        # The loop can stop here
                        break

            return result


    def __set_binding(self, dependency, service, reference):
        """
        Injects a service in the component

        :param dependency: The dependency handler
        :param service: The injected service
        :param reference: The reference of the injected service
        """
        with self._lock:
            # Set the value
            setattr(self.instance, dependency.field, dependency.get_value())

            # Call the component back
            self.__safe_callback(constants.IPOPO_CALLBACK_BIND,
                                 service, reference)

            self.__safe_field_callback(dependency.field,
                                       constants.IPOPO_CALLBACK_BIND_FIELD,
                                       service, reference)


    def __update_binding(self, dependency, service, reference, old_properties):
        """
        Calls back component binding and field binding methods when the
        properties of an injected dependency have been updated.
        
        :param dependency: The dependency handler
        :param service: The injected service
        :param reference: The reference of the injected service
        :param old_properties: Previous properties of the dependency
        """
        with self._lock:
            # Call the component back
            self.__safe_field_callback(dependency.field,
                                       constants.IPOPO_CALLBACK_UPDATE_FIELD,
                                       service, reference, old_properties)

            self.__safe_callback(constants.IPOPO_CALLBACK_UPDATE,
                                 service, reference, old_properties)


    def __unset_binding(self, dependency, service, reference):
        """
        Removes a service from the component

        :param dependency: The dependency handler
        :param service: The injected service
        :param reference: The reference of the injected service
        """
        with self._lock:
            # Call the component back
            self.__safe_field_callback(dependency.field,
                                       constants.IPOPO_CALLBACK_UNBIND_FIELD,
                                       service, reference)

            self.__safe_callback(constants.IPOPO_CALLBACK_UNBIND,
                                 service, reference)

            # Update the injected field
            setattr(self.instance, dependency.field, dependency.get_value())

            # Unget the service
            self.bundle_context.unget_service(reference)


# ------------------------------------------------------------------------------

def _set_factory_context(factory_class, bundle_context):
    """
    Transforms the context data dictionary into its FactoryContext object form.

    :param factory_class: A manipulated class
    :param bundle_context: The class bundle context
    :return: The factory context, None on error
    """
    if not hasattr(factory_class, constants.IPOPO_FACTORY_CONTEXT_DATA):
        # The class has not been manipulated, or too badly
        return None

    # Try to get the context dictionary (built using decorators)
    context_dict = getattr(factory_class, constants.IPOPO_FACTORY_CONTEXT_DATA)

    if not isinstance(context_dict, dict):
        # We got another form of context
        return None

    # Try to load the stored data
    try:
        context = FactoryContext.from_dictionary_form(context_dict)

    except (TypeError, ValueError):
        _logger.exception("Invalid data in manipulated class '%s'",
                          factory_class.__name__)
        # Work on the next class
        return None

    # Setup the context
    context.set_bundle_context(bundle_context)

    # Inject the constructed object
    setattr(factory_class, constants.IPOPO_FACTORY_CONTEXT, context)
    return context


def _load_bundle_factories(bundle):
    """
    Retrieves a list of pairs (FactoryContext, factory class) with all
    readable manipulated classes found in the bundle.

    :param bundle: A Bundle object
    :return: The list of factories loaded from the bundle
    """
    result = []

    # Get the Python module
    module = bundle.get_module()

    # Get the bundle context
    bundle_context = bundle.get_bundle_context()

    # Get all classes defined in the module
    for inspect_member in inspect.getmembers(module, inspect.isclass):

        # Get the class in the result tuple
        factory_class = inspect_member[1]

        context = _set_factory_context(factory_class, bundle_context)

        if context is None:
            # Error setting up the factory context
            continue

        result.append((context, factory_class))

    return result

# ------------------------------------------------------------------------------

def _field_property_generator(stored_instance):
    """
    Generates the methods called by the injected class properties

    :param stored_instance: A stored component instance
    """
    def get_value(self, name):
        """
        Retrieves the property value, from the iPOPO dictionaries

        :param name: The property name
        :return: The property value
        """
        assert stored_instance.context is not None
        return stored_instance.context.properties.get(name, None)


    def set_value(self, name, new_value):
        """
        Sets the property value and trigger an update event

        :param name: The property name
        :param new_value: The new property value
        """
        assert stored_instance.context is not None

        # Get the previous value
        old_value = stored_instance.context.properties.get(name, None)
        if new_value != old_value:
            # Change the property
            stored_instance.context.properties[name] = new_value

            # New value is different of the old one, trigger an event
            stored_instance.update_property(name, old_value, new_value)

        return new_value

    return (get_value, set_value)


def _field_controller_generator(stored_instance):
    """
    Generates the methods called by the injected controller

    :param stored_instance: A stored component instance
    """
    def get_value(self, name):
        """
        Retrieves the controller value, from the iPOPO dictionaries

        :param name: The property name
        :return: The property value
        """
        assert stored_instance.context is not None
        return stored_instance.get_controller_state(name)


    def set_value(self, name, new_value):
        """
        Sets the property value and trigger an update event

        :param name: The property name
        :param new_value: The new property value
        """
        assert stored_instance.context is not None

        # Get the previous value
        old_value = stored_instance.get_controller_state(name)
        if new_value != old_value:
            # Update the controller state
            stored_instance.set_controller_state(name, new_value)

        return new_value

    return (get_value, set_value)


def _manipulate_component(instance, stored_instance):
    """
    Manipulates the component instance to inject missing elements.

    Injects the properties handling
    """
    assert instance is not None
    assert isinstance(stored_instance, _StoredInstance)

    # Inject properties
    if stored_instance.context.factory_context.properties_fields:
        # Avoid injection of unused instance fields, avoiding more differences
        # between the class definition and the instance fields
        # -> Removes the overhead if the manipulated class has no __slots__ or
        #    when running in Pypy.
        getter, setter = _field_property_generator(stored_instance)

        # Prepare the methods names
        getter_name = "{0}{1}".format(constants.IPOPO_PROPERTY_PREFIX,
                                      constants.IPOPO_GETTER_SUFFIX)
        setter_name = "{0}{1}".format(constants.IPOPO_PROPERTY_PREFIX,
                                      constants.IPOPO_SETTER_SUFFIX)

        # Inject the getter and setter at the instance level
        setattr(instance, getter_name, getter)
        setattr(instance, setter_name, setter)

    # Inject controllers
    provides_tuples = stored_instance.context.factory_context.provides
    if provides_tuples:
        # Avoid injection of unused instance fields...
        controllers = set([value[1] for value in provides_tuples if value[1]])
        if controllers:
            # Controllers are valid by default
            for name in controllers:
                # Get the current value of the member (True by default)
                controller_value = getattr(instance, name, True)
                # Store the controller value
                stored_instance.set_controller_state(name, controller_value)

            # Prepare the methods names
            getter_name = "{0}{1}".format(constants.IPOPO_CONTROLLER_PREFIX,
                                          constants.IPOPO_GETTER_SUFFIX)
            setter_name = "{0}{1}".format(constants.IPOPO_CONTROLLER_PREFIX,
                                          constants.IPOPO_SETTER_SUFFIX)

            # Inject the getter and setter at the instance level
            getter, setter = _field_controller_generator(stored_instance)
            setattr(instance, getter_name, getter)
            setattr(instance, setter_name, setter)


# ------------------------------------------------------------------------------

class _IPopoService(object):
    """
    The iPOPO registry and service
    """

    def __init__(self, bundle_context):
        """
        Sets up the iPOPO registry

        :param bundle_context: The iPOPO bundle context
        """
        # Store the bundle context
        self.__context = bundle_context

        # Factories registry : name -> factory class
        self.__factories = {}

        # Instances registry : name -> _StoredInstance object
        self.__instances = {}

        # Event listeners
        self.__listeners = []

        # Auto-restarted components (Bundle -> [(factory, name, properties)]
        self.__auto_restart = {}

        # Service state
        self.running = False

        # Registries locks
        self.__factories_lock = threading.RLock()
        self.__instances_lock = threading.RLock()
        self.__listeners_lock = threading.RLock()


    def __get_stored_instances_by_factory(self, factory_name):
        """
        Retrieves the list of all stored instances objects corresponding to
        the given factory name

        :param factory_name: A factory name
        :return: All components instantiated from the given factory
        """
        with self.__instances_lock:
            return [stored_instance \
                    for stored_instance in self.__instances.values() \
                    if stored_instance.factory_name == factory_name]


    def _autorestart_store_components(self, bundle):
        """
        Stores the components of the given bundle with the auto-restart property
        
        :param bundle: A Bundle object
        """
        with self.__instances_lock:
            # Prepare the list of components
            store = self.__auto_restart.setdefault(bundle, [])

            for stored_instance in self.__instances.values():
                # Get the factory name
                factory = stored_instance.factory_name

                if self.get_factory_bundle(factory) is bundle:
                    # Factory from this bundle

                    # Test component properties
                    properties = stored_instance.context.properties
                    if properties.get(constants.IPOPO_AUTO_RESTART):
                        # Auto-restart property found
                        store.append((factory, stored_instance.name,
                                      properties))


    def _autorestart_components(self, bundle):
        """
        Restart the components of the given bundle
        
        :param bundle: A Bundle object
        """
        with self.__instances_lock:

            instances = self.__auto_restart.get(bundle)
            if not instances:
                # Nothing to do
                return

            for factory, name, properties in instances:
                try:
                    # Instantiate the given component
                    self.instantiate(factory, name, properties)

                except Exception as ex:
                    # Log error, but continue to work
                    _logger.exception("Error restarting component '%s' ('%s')"
                                      "from bundle %s (%d): %s", name, factory,
                                      bundle.get_symbolic_name(),
                                      bundle.get_bundle_id(), ex)


    def _autorestart_clear_components(self, bundle):
        """
        Clear the list of auto-restart components of the given bundle
        
        :param bundle: A Bundle object
        """
        with self.__instances_lock:
            # Simply delete the entry
            if bundle in self.__auto_restart:
                del self.__auto_restart[bundle]


    def _fire_ipopo_event(self, kind, factory_name, component_name=None):
        """
        Triggers an iPOPO event

        :param kind: Kind of event
        :param factory_name: Name of the factory associated to the event
        :param component_name: Name of the component instance associated to the
                               event
        """
        with self.__listeners_lock:
            # Use a copy of the list of listeners
            listeners = self.__listeners[:]

            for listener in listeners:
                try:
                    listener.handle_ipopo_event(constants.IPopoEvent(kind,
                                                             factory_name,
                                                             component_name))

                except:
                    _logger.exception("Error calling an iPOPO event handler")


    def _register_bundle_factories(self, bundle):
        """
        Registers all factories found in the given bundle

        :param bundle: A bundle
        """
        assert isinstance(bundle, Bundle)

        # Load the bundle factories
        factories = _load_bundle_factories(bundle)

        for context, factory_class in factories:
            # Register each found factory
            self._register_factory(context.name, factory_class, True)

            instances = getattr(factory_class, constants.IPOPO_INSTANCES, None)
            if isinstance(instances, dict):
                for name, properties in instances.items():
                    self.instantiate(context.name, name, properties)


    def _register_factory(self, factory_name, factory, override=True):
        """
        Registers a component factory

        :param factory_name: The name of the factory
        :param factory: The factory class object
        :param override: If true, previous factory is overridden, else an
                         exception is risen if a previous factory with that name
                         already exists
        :raise ValueError: The factory name already exists or is invalid
        :raise TypeError: Invalid factory type
        """
        if not factory_name or not is_string(factory_name):
            raise ValueError("A factory name must be a non-empty string")

        if not inspect.isclass(factory):
            raise TypeError("Invalid factory class '{0}'".format(
                                                        type(factory).__name__))

        with self.__factories_lock:
            if factory_name in self.__factories:
                if override:
                    _logger.info("Overriding factory '%s'", factory_name)

                else:
                    raise ValueError("'{0}' factory already exist".format(
                                                                factory_name))

            self.__factories[factory_name] = factory

            # Trigger an event
            self._fire_ipopo_event(constants.IPopoEvent.REGISTERED,
                                   factory_name)


    def _unregister_all_factories(self):
        """
        Unregisters all factories. This method should be called only after the
        iPOPO service has been unregistered (that's why it's not locked)
        """
        factories = list(self.__factories.keys())

        for factory_name in factories:
            self.unregister_factory(factory_name)


    def _unregister_bundle_factories(self, bundle):
        """
        Unregisters all factories of the given bundle

        :param bundle: A bundle
        """
        assert isinstance(bundle, Bundle)

        with self.__factories_lock:
            # Find out which factories must be removed
            to_remove = []

            for name in self.__factories:
                if self.get_factory_bundle(name) is bundle:
                    # Found
                    to_remove.append(name)

            # Remove all of them
            for factory in to_remove:
                self.unregister_factory(factory)


    def instantiate(self, factory_name, name, properties=None):
        """
        Instantiates a component from the given factory, with the given name

        :param factory_name: Name of the component factory
        :param name: Name of the instance to be started
        :return: The component instance
        :raise TypeError: The given factory is unknown
        :raise ValueError: The given name or factory name is invalid, or an
                           instance with the given name already exists
        :raise Exception: Something wrong occurred in the factory
        """
        # Test parameters
        if not factory_name or not is_string(factory_name):
            raise ValueError("Invalid factory name")

        if not name or not is_string(name):
            raise ValueError("Invalid component name")

        with self.__instances_lock:
            if name in self.__instances:
                raise ValueError("'{0}' is an already running instance name" \
                                 .format(name))

            with self.__factories_lock:
                # Can raise a ValueError exception
                factory = self.__factories.get(factory_name, None)
                if factory is None:
                    raise TypeError("Unknown factory '{0}'" \
                                    .format(factory_name))

                # Get the factory context
                factory_context = getattr(factory, \
                                          constants.IPOPO_FACTORY_CONTEXT, None)
                if factory_context is None:
                    raise TypeError("Factory context missing in '{0}'" \
                                    .format(factory_name))

            # Create component instance
            try:
                instance = factory()

            except:
                _logger.exception("Error creating the instance '%s' " \
                                  "from factory '%s'", name, factory_name)

                raise TypeError("Factory '{0}' failed to create '{1}'" \
                                .format(factory_name, name))

            # Normalize given properties
            if properties is None or not isinstance(properties, dict):
                properties = {}


            # Use framework properties to fill missing ones
            framework = self.__context.get_bundle(0)
            for property_name in factory_context.properties:
                if property_name not in properties:
                    # Missing property
                    value = framework.get_property(property_name)
                    if value is not None:
                        # Set the property value
                        properties[property_name] = value

            # Set the instance context
            component_context = ComponentContext(factory_context, name, \
                                                 properties)

            # Prepare the stored instance
            stored_instance = _StoredInstance(self, component_context, instance)

            # Manipulate the properties
            _manipulate_component(instance, stored_instance)

            # Store the instance
            self.__instances[name] = stored_instance

        # Start the manager
        stored_instance.start()

        # Notify listeners now that every thing is ready to run
        self._fire_ipopo_event(constants.IPopoEvent.INSTANTIATED,
                               factory_name, name)

        # Try to validate it
        stored_instance.update_bindings()
        stored_instance.check_lifecycle()

        return instance


    def invalidate(self, name):
        """
        Invalidates the given component

        :param name: Name of the component to invalidate
        :raise ValueError: Invalid component name
        """
        with self.__instances_lock:
            if name not in self.__instances:
                raise ValueError("Unknown component instance '{0}'"\
                                 .format(name))

            stored_instance = self.__instances[name]

            # Call back the component during the invalidation
            stored_instance.invalidate(True)


    def is_registered_factory(self, name):
        """
        Tests if the given name is in the factory registry

        :param name: A factory name to be tested
        """
        with self.__factories_lock:
            return name in self.__factories


    def is_registered_instance(self, name):
        """
        Tests if the given name is in the instance registry

        :param name: A component name to be tested
        """
        with self.__instances_lock:
            return name in self.__instances


    def kill(self, name):
        """
        Kills the given component

        :param name: Name of the component to kill
        :raise ValueError: Invalid component name
        """
        if not name:
            raise ValueError("Name can't be None or empty")

        with self.__instances_lock:
            if name not in self.__instances:
                raise ValueError("Unknown component instance '{0}'" \
                                 .format(name))

            stored_instance = self.__instances.pop(name)

            # Kill it
            stored_instance.kill()


    def register_factory(self, bundle_context, factory):
        """
        Registers a manually created factory, using decorators programmatically

        :param bundle_context: The factory bundle context
        :param factory: A manipulated class
        :return: True if the factory has been registered
        :raise ValueError: Invalid parameter, or factory already registered
        :raise TypeError: Invalid factory type (not a manipulated class)
        """
        if factory is None or bundle_context is None:
            # Invalid parameter, to nothing
            raise ValueError("Invalid parameter")

        context = _set_factory_context(factory, bundle_context)
        if not context:
            raise TypeError("Not a manipulated class (no context found)")

        self._register_factory(context.name, factory, False)
        return True


    def unregister_factory(self, factory_name):
        """
        Unregisters the given component factory

        :param factory_name: Name of the factory to unregister
        :return: True the factory has been removed, False if the factory is
                 unknown
        """
        if not factory_name or not is_string(factory_name):
            # Invalid name
            return False

        with self.__factories_lock:
            if factory_name not in self.__factories:
                # Unknown factory
                return False

            # Trigger an event
            self._fire_ipopo_event(constants.IPopoEvent.UNREGISTERED,
                                   factory_name)

            # Invalidate and delete all components of this factory
            with self.__instances_lock:
                # Compute the list of __instances to remove
                to_remove = self.__get_stored_instances_by_factory(factory_name)

                # Remove instances from the registry: avoids dependencies \
                # update to link against a component from this factory again.
                for instance in to_remove:
                    self.kill(instance.name)

            # Remove the factory from the registry
            del self.__factories[factory_name]

        return True


    def add_listener(self, listener):
        """
        Register an iPOPO event listener.

        The event listener must have a method with the following prototype :

        .. python::

           def handle_ipopo_event(self, event):
               '''
               :param event: A IPopoEvent object
               '''
               # ...

        :param listener: The listener to register
        :return: True if the listener has been added to the registry
        """
        with self.__listeners_lock:
            return add_listener(self.__listeners, listener)


    def remove_listener(self, listener):
        """
        Unregister an iPOPO event listener.

        :param listener: The listener to register
        :return: True if the listener has been removed from the registry
        """
        with self.__listeners_lock:
            return remove_listener(self.__listeners, listener)


    def get_instances(self):
        """
        Retrieves the list of the currently registered component instances
        
        :return: A list of (name, factory name, state) tuples.
        """
        with self.__instances_lock:
            result = []
            for name, stored_instance in self.__instances.items():
                result.append((name, stored_instance.factory_name,
                               stored_instance.state))

            result.sort()
            return result


    def get_instance_details(self, name):
        """
        Retrieves a snapshot of the given component instance.
        The result dictionary has the following keys:
        
        * name: The component name
        * factory: The name of the component factory
        * state: The current component state
        * service (optional): The reference of the service provided by the
          component
        * dependencies: A dictionary associating field names with the following
          dictionary:
        
          * handler: The name of the type of the dependency handler
          * filter (optional): The requirement LDAP filter
          * optional: A flag indicating whether the requirement is optional or
            not
          * aggregate: A flag indicating whether the requirement is a set of
            services or not
          * binding: A list of the ServiceReference the component is bound to
        
        :param name: The name of a component instance
        :return: A dictionary of details
        :raise ValueError: Invalid component name
        """
        if not is_string(name):
            raise ValueError("Component name must be a string")

        with self.__instances_lock:
            if name not in self.__instances:
                raise ValueError("Unknown component: {0}".format(name))

            stored_instance = self.__instances[name]
            assert isinstance(stored_instance, _StoredInstance)
            with stored_instance._lock:
                result = {}
                result["name"] = stored_instance.name

                # Factory name
                result["factory"] = stored_instance.factory_name

                # Component state
                result["state"] = stored_instance.state

                # Provided service
                result["services"] = {}
                for handler in stored_instance.get_handlers(
                                            constants.HANDLER_SERVICE_PROVIDER):
                    svc_ref = handler.get_service_reference()
                    if svc_ref is not None:
                        svc_id = svc_ref.get_property(pelix.SERVICE_ID)
                        result["services"][svc_id] = svc_ref

                # Dependencies
                result["dependencies"] = {}
                for dependency in stored_instance.get_handlers(
                                            constants.HANDLER_DEPENDENCY):
                    # Dependency
                    info = result["dependencies"][dependency.field] = {}
                    info["handler"] = type(dependency).__name__

                    # Requirement
                    req = dependency.requirement
                    if req.filter:
                        info["filter"] = str(req.filter)

                    info["optional"] = req.optional
                    info["aggregate"] = req.aggregate

                    # Bindings
                    info["bindings"] = dependency.get_bindings()

                # All done
                return result


    def get_factories(self):
        """
        Retrieves the names of the registered factories

        :return: A list of factories. Can be empty.
        """
        with self.__factories_lock:
            result = list(self.__factories.keys())
            result.sort()
            return result


    def get_factory_bundle(self, name):
        """
        Retrieves the Pelix Bundle object that registered the given factory
        
        :param name: The name of a factory
        :return: The Bundle that registered the given factory
        :raise ValueError: Invalid factory
        """
        with self.__factories_lock:
            if name not in self.__factories:
                raise ValueError("Unknown factory '{0}'".format(name))

            factory = self.__factories[name]

            # Bundle Context is stored in the Factory Context
            factory_context = getattr(factory, constants.IPOPO_FACTORY_CONTEXT)
            return factory_context.bundle_context.get_bundle()

# ------------------------------------------------------------------------------

class _IPopoActivator(object):
    """
    The iPOPO bundle activator for Pelix
    """

    def __init__(self):
        """
        Sets up the activator
        """
        self._registration = None
        self.service = None


    def start(self, context):
        """
        The bundle has started

        :param context: The bundle context
        """
        assert isinstance(context, BundleContext)

        # Register the iPOPO service
        self.service = _IPopoService(context)
        self._registration = context.register_service(\
                                        constants.IPOPO_SERVICE_SPECIFICATION, \
                                        self.service, {})

        # Register as a bundle listener
        context.add_bundle_listener(self)

        # Register as a framework stop listener
        context.add_framework_stop_listener(self)

        # Service enters in "run" mode
        self.service.running = True

        # Get all factories
        for bundle in context.get_bundles():
            if bundle.get_state() == Bundle.ACTIVE:
                # Bundle is active, register its factories
                self.service._register_bundle_factories(bundle)


    def stop(self, context):
        """
        The bundle has stopped

        :param context: The bundle context
        """
        assert isinstance(context, BundleContext)

        # The service is not in the "run" mode anymore
        self.service.running = False

        # Unregister the listener
        context.remove_bundle_listener(self)

        # Unregister the framework stop listener
        context.remove_framework_stop_listener(self)

        # Unregister the iPOPO service
        self._registration.unregister()

        # Clean up the service
        self.service._unregister_all_factories()


    def bundle_changed(self, event):
        """
        A bundle event has been triggered

        :param event: The bundle event
        """
        assert isinstance(event, BundleEvent)

        kind = event.get_kind()
        bundle = event.get_bundle()

        if kind == BundleEvent.STOPPING_PRECLEAN:
            # A bundle is gone, remove its factories after the deactivator has
            # been called. That way, the deactivator can kill manually started
            # components.
            self.service._unregister_bundle_factories(bundle)

        elif kind == BundleEvent.STARTING:
            # A bundle is staring, register its factories before its activator
            # is called. That way, the activator can use the registered
            # factories.
            self.service._register_bundle_factories(bundle)

        elif kind == BundleEvent.UPDATE_BEGIN:
            # A bundle will be updated, store its auto-restart component
            self.service._autorestart_store_components(bundle)

        elif kind == BundleEvent.UPDATED:
            # Update has finished, restart stored components
            self.service._autorestart_components(bundle)
            self.service._autorestart_clear_components(bundle)

        elif kind == BundleEvent.UPDATE_FAILED:
            # Update failed, clean the stored components
            self.service._autorestart_clear_components(bundle)


    def framework_stopping(self):
        """
        Called when the framework is stopping
        """
        # Avoid new injections, as all bundles will be stopped
        self.service.running = False

# ------------------------------------------------------------------------------

# The activator instance
activator = _IPopoActivator()



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: XML-RPC implementation

Based on standard package xmlrpclib

:author: Thomas Calmant
:copyright: Copyright 2013, isandlaTech
:license: GPLv3
:version: 0.1
:status: Alpha

..

    This file is part of iPOPO.

    iPOPO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    iPOPO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with iPOPO. If not, see <http://www.gnu.org/licenses/>.

**TODO:**
* "system" methods (list, help, ...)
"""

# Module version
__version_info__ = (0, 1, 0)
__version__ = ".".join(map(str, __version_info__))

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

# iPOPO decorators
from pelix.ipopo.decorators import ComponentFactory, Requires, Validate, \
    Invalidate, Property, Provides

# Pelix constants
import pelix.http
import pelix.remote

# Standard library
import logging
import sys
import uuid

# XML RPC modules
if sys.version_info[0] == 3:
    # Python 3
    from xmlrpc.server import SimpleXMLRPCDispatcher
    import xmlrpc.client as xmlrpclib

else:
    # Python 2
    from SimpleXMLRPCServer import SimpleXMLRPCDispatcher
    import xmlrpclib

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------

class _XmlRpcServlet(SimpleXMLRPCDispatcher):
    """
    A XML-RPC servlet that can be registered in the Pelix HTTP service
    
    Calls the dispatch method given in the constructor
    """
    def __init__(self, dispatch_method, encoding=None):
        """
        Sets up the servlet
        """
        SimpleXMLRPCDispatcher.__init__(self, allow_none=True,
                                        encoding=encoding)

        # Register the system.* functions
        self.register_introspection_functions()

        # Make a link to the dispatch method
        self._dispatch_method = dispatch_method


    def _simple_dispatch(self, name, params):
        """
        Dispatch method
        """
        try:
            # Internal method
            return self.funcs[name](*params)

        except KeyError:
            # Other method
            return self._dispatch_method(name, params)


    def do_POST(self, request, response):
        """
        Handles a HTTP POST request
        
        :param request: The HTTP request bean
        :param request: The HTTP response handler
        """
        # Get the request content
        data = request.read_data()

        # Dispatch
        result = self._marshaled_dispatch(data, self._simple_dispatch)

        # Send the result
        response.send_content(200, result, 'text/xml')

# ------------------------------------------------------------------------------

@ComponentFactory("pelix-xmlrpc-exporter-factory")
@Requires('_dispatcher', pelix.remote.SERVICE_DISPATCHER)
@Requires('_http', pelix.http.HTTP_SERVICE)
@Property('_path', pelix.http.HTTP_SERVLET_PATH, '/XML-RPC')
@Property('_kind', 'endpoints.kind', 'xmlrpc')
class XmlRpcServiceExporter(object):
    """
    XML-RPC Remote Services exporter
    """
    def __init__(self):
        """
        Sets up the exporter
        """
        # Bundle context
        self._context = None

        # Dispatcher
        self._dispatcher = None
        self._kind = None

        # HTTP Service
        self._http = None
        self._path = None

        # XML-RPC servlet
        self._servlet = None

        # Exported services: Name -> ExportEndpoint
        self.__endpoints = {}

        # Service Reference -> ExportEndpoint
        self.__registrations = {}


    def _dispatch(self, method, params):
        """
        Called by the XML-RPC servlet: calls the method of an exported service
        """
        # Get the best matching name
        matching = None
        len_found = 0
        for name in self.__endpoints:
            if len(name) > len_found and method.startswith(name + "."):
                # Better matching end point name (longer that previous one)
                matching = name
                len_found = len(matching)

        if matching is None:
            # No end point name match
            raise KeyError("No end point found for: {0}".format(method))

        # Extract the method name. (+1 for the trailing dot)
        method_name = method[len_found + 1:]

        # Call the dispatcher
        return self._dispatcher.dispatch(self._kind, matching,
                                         method_name, params)


    def _compute_endpoint_name(self, reference):
        """
        Computes the end point name according to service properties
        
        :param reference: A ServiceReference object
        :return: The computed end point name
        """
        service_id = reference.get_property(pelix.framework.SERVICE_ID)
        endpoint_name = reference.get_property(pelix.remote.PROP_ENDPOINT_NAME)
        if not endpoint_name:
            endpoint_name = 'service_{0}'.format(service_id)

        return endpoint_name


    def _export_service(self, reference):
        """
        Exports the given service
        
        :param reference: A ServiceReference object
        :return: True if the service has been exported, else False
        """
        # Compute the end point name
        endpoint_name = self._compute_endpoint_name(reference)
        if endpoint_name in self.__endpoints:
            # Already known end point
            _logger.error("Already known end point %s for XML-RPC",
                          endpoint_name)
            return False

        # Get the service
        try:
            service = self._context.get_service(reference)
            if service is None:
                _logger.error("Invalid service for reference %s",
                              str(reference))

        except pelix.framework.BundleException as ex:
            _logger.error("Error retrieving the service to export: %s", ex)
            return False

        # Create the registration information
        try:
            endpoint = pelix.remote.ExportEndpoint(str(uuid.uuid4()),
                                                   self._kind, endpoint_name,
                                                   reference, service,
                                                   self.get_access())
        except ValueError:
            # Invalid end point
            return False

        try:
            # Register the end point
            self._dispatcher.add_endpoint(self._kind, endpoint_name, endpoint)

        except KeyError as ex:
            _logger.error("Error registering end point: %s", ex)

        else:
            # Store informations
            self.__endpoints[endpoint_name] = endpoint
            self.__registrations[reference] = endpoint
            return True

        return False


    def _update_service(self, reference, old_properties):
        """
        Service properties updated
        """
        # Compute the new end point name
        new_name = self._compute_endpoint_name(reference)

        # Get the end point
        endpoint = self.__registrations[reference]
        if endpoint.name != new_name:
            # Name changed -> re-export the service
            self._unexport_service(reference)
            self._export_service(reference)

        else:
            # Notify the dispatcher
            self._dispatcher.update_endpoint(self._kind, endpoint.name,
                                             endpoint, old_properties)


    def _unexport_service(self, reference):
        """
        Stops the export of the given service
        
        :param reference: A ServiceReference object
        """
        # Find the corresponding end point
        endpoint = self.__registrations.get(reference)
        if endpoint is not None:
            # Delete the registration
            del self.__registrations[reference]
            del self.__endpoints[endpoint.name]

            # Unregister the service from the dispatcher
            self._dispatcher.remove_endpoint(self._kind, endpoint.name)


    def service_changed(self, event):
        """
        Called when a service event is triggered
        """
        kind = event.get_kind()
        svcref = event.get_service_reference()

        if kind == pelix.framework.ServiceEvent.REGISTERED:
            # Simply export the service
            self._export_service(svcref)

        elif kind == pelix.framework.ServiceEvent.MODIFIED:
            # Matching registering or updated service
            if svcref not in self.__registrations:
                # New match
                self._export_service(svcref)

            else:
                # Properties modification:
                # Re-export if endpoint.name has changed
                self._update_service(svcref, event.get_previous_properties())

        elif svcref in self.__registrations and \
                (kind == pelix.framework.ServiceEvent.UNREGISTERING or \
                 kind == pelix.framework.ServiceEvent.MODIFIED_ENDMATCH):
            # Service is updated or unregistering
            self._unexport_service(svcref)


    def get_access(self):
        """
        Retrieves the URL to access this component
        """
        port = self._http.get_access()[1]
        return "http://{{server}}:{0}{1}".format(port, self._path)


    @Validate
    def validate(self, context):
        """
        Component validated
        """
        # Store the context
        self._context = context

        # Prepare the service filter
        ldapfilter = '(|(|({0}=xmlrpc)({0}=\*))(&(!({0}=*))({1}=*)))' \
                    .format(pelix.remote.PROP_EXPORTED_CONFIGS,
                            pelix.remote.PROP_EXPORTED_INTERFACES)

        # Export existing services
        existing_ref = self._context.get_all_service_references(None,
                                                                ldapfilter)
        if existing_ref is not None:
            for reference in existing_ref:
                self._export_service(reference)

        # Register a service listener, to update the exported services state
        self._context.add_service_listener(self, ldapfilter)

        # Create/register the servlet
        self._servlet = _XmlRpcServlet(self._dispatch)
        self._http.register_servlet(self._path, self._servlet)


    @Invalidate
    def invalidate(self, context):
        """
        Component invalidated
        """
        # Unregister the service listener
        context.remove_service_listener(self)

        # Unregister the servlet
        self._http.unregister(None, self._servlet)

        # Remove all exports
        for reference in self.__registrations.keys():
            self._unexport_service(reference)

        # Clean up the storage
        self.__endpoints.clear()
        self.__registrations.clear()

        # Clean up members
        self._servlet = None
        self._context = None

# ------------------------------------------------------------------------------

class _ServiceCallProxy(object):
    """
    Service call proxy
    """
    def __init__(self, name, url):
        """
        Sets up the call proxy
        
        :param name: End point name
        :param url: End point URL
        """
        self.__name = name
        self.__url = url


    def __getattr__(self, name):
        """
        Prefixes the requested attribute name by the endpoint name
        """
        # Make a proxy for this call
        # This is an ugly trick to handle multithreaded calls, as the underlying
        # proxy re-uses the same connection when possible: sometimes it means
        # sending a request before retrieving a result
        proxy = xmlrpclib.ServerProxy(self.__url, allow_none=True)
        return getattr(proxy, "{0}.{1}".format(self.__name, name))


@ComponentFactory("pelix-xmlrpc-importer-factory")
@Provides(pelix.remote.SERVICE_ENDPOINT_LISTENER)
@Property('_kind', 'endpoints.kind', 'xmlrpc')
@Property('_listener_flag', pelix.remote.PROP_LISTEN_IMPORTED, True)
class XmlRpcServiceImporter(object):
    """
    XML-RPC Remote Services importer
    """
    def __init__(self):
        """
        Sets up the exporter
        """
        # Bundle context
        self._context = None

        # Component properties
        self._kind = None
        self._listener_flag = True

        # Registered services (end point -> reference)
        self.__registrations = {}


    def endpoint_added(self, endpoint):
        """
        An end point has been imported
        """
        if endpoint.kind != self._kind and endpoint.kind != '*':
            # Not for us
            return

        # Register the service
        svc = _ServiceCallProxy(endpoint.name, endpoint.url)
        svc_reg = self._context.register_service(endpoint.specifications, svc,
                                                 endpoint.properties)

        # Store references
        self.__registrations[endpoint.uid] = svc_reg


    def endpoint_updated(self, endpoint, old_properties):
        """
        An end point has been updated
        """
        if endpoint.uid not in self.__registrations:
            # Unknown end point
            return

        # Update service properties
        svc_reg = self.__registrations[endpoint.uid]
        svc_reg.set_properties(endpoint.properties)


    def endpoint_removed(self, endpoint):
        """
        An end point has been removed
        """
        if endpoint.uid not in self.__registrations:
            # Unknown end point
            return

        # Pop references
        svc_reg = self.__registrations.pop(endpoint.uid)

        # Unregister the service
        svc_reg.unregister()


    @Validate
    def validate(self, context):
        """
        Component validated
        """
        self._context = context


    @Invalidate
    def invalidate(self, context):
        """
        Component invalidated
        """
        self._context = None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix OSGi-like services packages

:author: Thomas Calmant
:copyright: Copyright 2013, isandlaTech
:license: GPLv3
:version: 0.1
:status: Alpha

..

    This file is part of iPOPO.

    iPOPO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    iPOPO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with iPOPO. If not, see <http://www.gnu.org/licenses/>.
"""

# Module version
__version_info__ = (0, 1, 0)
__version__ = ".".join(map(str, __version_info__))

# Documentation strings format
__docformat__ = "restructuredtext en"

#-------------------------------------------------------------------------------

SERVICE_EVENT_ADMIN = "pelix.services.eventadmin"
""" Specification of the EventAdmin service """

SERVICE_EVENT_HANDLER = "pelix.services.eventadmin.handler"
""" Specification of an EventAdmin event handler """

PROP_EVENT_TOPICS = "event.topics"
""" **List** of the topics handled by an event handler """

PROP_EVENT_FILTER = "event.filter"
""" Filter on events properties for an event handler """

EVENT_PROP_FRAMEWORK_UID = "event.sender.framework.uid"
""" UID of the framework that emitted the event """

EVENT_PROP_TIMESTAMP = "event.timestamp"
""" Time stamp of the event, compute during the call of send() or post() """
